import logging
import sys
from functools import wraps
from timeit import default_timer

logger = logging.getLogger("truefoundry.workflow.remote_filesystem")


def init_logger(level=logging.WARNING):
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(level)
    formatter = logging.Formatter(
        "[%(name)s] %(asctime)s %(levelname)s %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False


def log_time(prefix: str = ""):
    """Decorator to log the time taken by I/O operations."""

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = default_timer()
            result = func(*args, **kwargs)
            elapsed_time = default_timer() - start_time
            logger.info(f"{prefix}{func.__name__} took {elapsed_time:.2f} seconds")
            return result

        return wrapper

    return decorator



import os
import posixpath
import sys
import tempfile
import uuid
from concurrent.futures import FIRST_EXCEPTION, Future, ThreadPoolExecutor, wait
from shutil import rmtree
from threading import Event
from typing import Any, Callable, Dict, Iterator, List, Optional, Sequence, Tuple
from urllib.parse import unquote
from urllib.request import pathname2url

from rich.console import _is_jupyter
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from tqdm.utils import CallbackIOWrapper
from truefoundry_sdk import (
    FileInfo,
    MultiPartUploadResponse,
    MultiPartUploadStorageProvider,
    Operation,
    SignedUrl,
)

from truefoundry import client
from truefoundry.common.constants import ENV_VARS
from truefoundry.common.request_utils import (
    augmented_raise_for_status,
    cloud_storage_http_request,
)
from truefoundry.common.storage_provider_utils import (
    MultiPartUpload,
    _FileMultiPartInfo,
    azure_multi_part_upload,
    decide_file_parts,
    s3_compatible_multipart_upload,
    truncate_path_for_progress,
)
from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    ApiClient,
    MlfoundryArtifactsApi,
    RunArtifactsApi,
)
from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.logger import logger
from truefoundry.ml.session import _get_api_client
from truefoundry.pydantic_v1 import BaseModel, root_validator

_LIST_FILES_PAGE_SIZE = 500
_GENERATE_SIGNED_URL_BATCH_SIZE = 50
DEFAULT_PRESIGNED_URL_EXPIRY_TIME = 3600


def _can_display_progress(user_choice: Optional[bool] = None) -> bool:
    if user_choice is False:
        return False

    if sys.stdout.isatty():
        return True
    elif _is_jupyter():
        try:
            from IPython.display import display  # noqa: F401
            from ipywidgets import Output  # noqa: F401

            return True
        except ImportError:
            logger.warning(
                "Detected Jupyter Environment. Install `ipywidgets` to display live progress bars.",
            )
    if user_choice is True:
        logger.warning(
            "`progress` argument is set to True but did not detect tty "
            "or jupyter environment with ipywidgets installed. "
            "Progress bars may not be displayed. "
        )
        return True
    return False


def relative_path_to_artifact_path(path):
    if os.path == posixpath:
        return path
    if os.path.abspath(path) == path:
        raise Exception("This method only works with relative paths.")
    return unquote(pathname2url(path))


def bad_path_message(name):
    return (
        "Names may be treated as files in certain cases, and must not resolve to other names"
        " when treated as such. This name would resolve to '%s'"
    ) % posixpath.normpath(name)


def path_not_unique(name):
    norm = posixpath.normpath(name)
    return norm != name or norm == "." or norm.startswith("..") or norm.startswith("/")


def verify_artifact_path(artifact_path):
    if artifact_path and path_not_unique(artifact_path):
        raise MlFoundryException(
            f"Invalid artifact path: {artifact_path!r}. {bad_path_message(artifact_path)!r}"
        )


def _signed_url_upload_file(
    signed_url: SignedUrl,
    local_file: str,
    progress_bar: Progress,
    abort_event: Optional[Event] = None,
):
    if os.stat(local_file).st_size == 0:
        with cloud_storage_http_request(
            method="put",
            url=signed_url.signed_url,
            data="",
            exception_class=MlFoundryException,  # type: ignore
        ) as response:
            augmented_raise_for_status(response, exception_class=MlFoundryException)  # type: ignore
        return

    task_id = progress_bar.add_task(
        f"[green]⬆ {truncate_path_for_progress(local_file, 64, relpath=True)}",
        start=True,
        visible=True,
    )

    def callback(length):
        progress_bar.update(task_id, advance=length, total=os.stat(local_file).st_size)
        if abort_event and abort_event.is_set():
            raise Exception("aborting upload")

    with open(local_file, "rb") as file:
        # NOTE: Azure Put Blob does not support Transfer Encoding header.
        wrapped_file = CallbackIOWrapper(callback, file, "read")
        with cloud_storage_http_request(
            method="put",
            url=signed_url.signed_url,
            data=wrapped_file,
            exception_class=MlFoundryException,  # type: ignore
        ) as response:
            augmented_raise_for_status(response, exception_class=MlFoundryException)  # type: ignore

    if progress_bar is not None:
        progress_bar.refresh()


def _download_file_using_http_uri(
    http_uri,
    download_path,
    chunk_size=ENV_VARS.TFY_ARTIFACTS_DOWNLOAD_CHUNK_SIZE_BYTES,
    callback: Optional[Callable[[int, int], Any]] = None,
):
    """
    Downloads a file specified using the `http_uri` to a local `download_path`. This function
    uses a `chunk_size` to ensure an OOM error is not raised a large file is downloaded.
    Note : This function is meant to download files using presigned urls from various cloud
            providers.
    """
    with cloud_storage_http_request(
        method="get",
        url=http_uri,
        stream=True,
        exception_class=MlFoundryException,  # type: ignore
    ) as response:
        augmented_raise_for_status(response, exception_class=MlFoundryException)  # type: ignore
        file_size = int(response.headers.get("Content-Length", -1))
        if file_size == 0 and callback:
            # special case for empty files
            callback(1, 1)
        file_size = file_size if file_size > 0 else 0
        with open(download_path, "wb") as output_file:
            for chunk in response.iter_content(chunk_size=chunk_size):
                if callback:
                    callback(len(chunk), file_size)
                if not chunk:
                    break
                output_file.write(chunk)
                if ENV_VARS.TFY_ARTIFACTS_DOWNLOAD_FSYNC_CHUNKS:
                    output_file.flush()
                    os.fsync(output_file.fileno())


def _any_future_has_failed(futures) -> bool:
    return any(
        future.done() and not future.cancelled() and future.exception() is not None
        for future in futures
    )


class ArtifactIdentifier(BaseModel):
    artifact_version_id: Optional[uuid.UUID] = None
    dataset_id: Optional[str] = None
    dataset_fqn: Optional[str] = None

    @root_validator
    def _check_identifier_type(cls, values: Dict[str, Any]):
        if not values.get("artifact_version_id", False) and not values.get(
            "dataset_id", False
        ):
            raise MlFoundryException(
                "One of the version_id or dataset_id should be passed"
            )
        if values.get("artifact_version_id", False) and values.get("dataset_id", False):
            raise MlFoundryException(
                "Exactly one of version_id or dataset_id should be passed"
            )
        return values


class MlFoundryArtifactsRepository:
    def __init__(
        self,
        artifact_identifier: ArtifactIdentifier,
        api_client: Optional[ApiClient] = None,
    ):
        self.artifact_identifier = artifact_identifier
        self._api_client = api_client or _get_api_client()
        self._run_artifacts_api = RunArtifactsApi(api_client=self._api_client)
        self._mlfoundry_artifacts_api = MlfoundryArtifactsApi(
            api_client=self._api_client
        )

    # TODO (chiragjn): Refactor these methods - if else is very inconvenient
    def get_signed_urls_for_read(
        self,
        artifact_identifier: ArtifactIdentifier,
        paths,
    ) -> List[SignedUrl]:
        if artifact_identifier.artifact_version_id:
            signed_urls_response = client.artifact_versions.get_signed_urls(
                id=str(artifact_identifier.artifact_version_id),
                paths=paths,
                operation=Operation.READ,
            )
            signed_urls = signed_urls_response.data
        elif artifact_identifier.dataset_id:
            signed_urls_dataset_response = client.data_directories.get_signed_urls(
                id=str(artifact_identifier.dataset_id),
                paths=paths,
                operation=Operation.READ,
            )
            signed_urls = signed_urls_dataset_response.data

        else:
            raise ValueError(
                "Invalid artifact type - both `artifact_version_id` and `dataset_id` both are None"
            )
        return signed_urls

    def get_signed_urls_for_write(
        self,
        artifact_identifier: ArtifactIdentifier,
        paths: List[str],
    ) -> List[SignedUrl]:
        if artifact_identifier.artifact_version_id:
            signed_urls_response = client.artifact_versions.get_signed_urls(
                id=str(artifact_identifier.artifact_version_id),
                paths=paths,
                operation=Operation.WRITE,
            )
            signed_urls = signed_urls_response.data
        elif artifact_identifier.dataset_id:
            signed_urls_dataset_response = client.data_directories.get_signed_urls(
                id=str(artifact_identifier.dataset_id),
                paths=paths,
                operation=Operation.WRITE,
            )
            signed_urls = signed_urls_dataset_response.data
        else:
            raise ValueError(
                "Invalid artifact type - both `artifact_version_id` and `dataset_id` both are None"
            )
        return signed_urls

    def _normal_upload(
        self,
        local_file: str,
        artifact_path: str,
        signed_url: Optional[SignedUrl],
        progress_bar: Progress,
        abort_event: Optional[Event] = None,
    ):
        if not signed_url:
            signed_url = self.get_signed_urls_for_write(
                artifact_identifier=self.artifact_identifier, paths=[artifact_path]
            )[0]

        if progress_bar.disable:
            logger.info(
                "Uploading %s to %s",
                local_file,
                artifact_path,
            )

        _signed_url_upload_file(
            signed_url=signed_url,
            local_file=local_file,
            abort_event=abort_event,
            progress_bar=progress_bar,
        )
        logger.debug("Uploaded %s to %s", local_file, artifact_path)

    def _create_multipart_upload_for_identifier(
        self,
        artifact_identifier: ArtifactIdentifier,
        path,
        num_parts,
    ) -> MultiPartUpload:
        if artifact_identifier.artifact_version_id:
            create_multipart_response: MultiPartUploadResponse = (
                client.artifact_versions.create_multi_part_upload(
                    id=str(artifact_identifier.artifact_version_id),
                    path=path,
                    num_parts=num_parts,
                )
            )
            multipart_upload = create_multipart_response.data

        elif artifact_identifier.dataset_id:
            create_multipart_for_dataset_response: MultiPartUploadResponse = (
                client.data_directories.create_multipart_upload(
                    id=str(artifact_identifier.dataset_id),
                    path=path,
                    num_parts=num_parts,
                )
            )
            multipart_upload = create_multipart_for_dataset_response.data

        else:
            raise ValueError(
                "Invalid artifact type - both `artifact_version_id` and `dataset_id` both are None"
            )
        return multipart_upload

    def _multipart_upload(
        self,
        local_file: str,
        artifact_path: str,
        multipart_info: _FileMultiPartInfo,
        executor: ThreadPoolExecutor,
        progress_bar: Progress,
        abort_event: Optional[Event] = None,
    ):
        if progress_bar.disable:
            logger.info(
                "Uploading %s to %s using multipart upload",
                local_file,
                artifact_path,
            )

        multipart_upload = self._create_multipart_upload_for_identifier(
            artifact_identifier=self.artifact_identifier,
            path=artifact_path,
            num_parts=multipart_info.num_parts,
        )
        if (
            multipart_upload.storage_provider
            is MultiPartUploadStorageProvider.S3COMPATIBLE
        ):
            s3_compatible_multipart_upload(
                multipart_upload=MultiPartUpload.parse_obj(multipart_upload.dict()),
                local_file=local_file,
                executor=executor,
                multipart_info=multipart_info,
                abort_event=abort_event,
                progress_bar=progress_bar,
                exception_class=MlFoundryException,  # type: ignore
            )
        elif (
            multipart_upload.storage_provider
            is MultiPartUploadStorageProvider.AZURE_BLOB
        ):
            azure_multi_part_upload(
                multipart_upload=MultiPartUpload.parse_obj(multipart_upload.dict()),
                local_file=local_file,
                executor=executor,
                multipart_info=multipart_info,
                abort_event=abort_event,
                progress_bar=progress_bar,
                exception_class=MlFoundryException,  # type: ignore
            )
        else:
            raise NotImplementedError()

    def _upload_file(
        self,
        local_file: str,
        artifact_path: str,
        multipart_info: _FileMultiPartInfo,
        progress_bar: Progress,
        signed_url: Optional[SignedUrl] = None,
        abort_event: Optional[Event] = None,
        executor_for_multipart_upload: Optional[ThreadPoolExecutor] = None,
    ):
        if multipart_info.num_parts == 1:
            return self._normal_upload(
                local_file=local_file,
                artifact_path=artifact_path,
                signed_url=signed_url,
                abort_event=abort_event,
                progress_bar=progress_bar,
            )

        if not executor_for_multipart_upload:
            with ThreadPoolExecutor(
                max_workers=ENV_VARS.TFY_ARTIFACTS_UPLOAD_MAX_WORKERS
            ) as executor:
                return self._multipart_upload(
                    local_file=local_file,
                    artifact_path=artifact_path,
                    executor=executor,
                    multipart_info=multipart_info,
                    progress_bar=progress_bar,
                )

        return self._multipart_upload(
            local_file=local_file,
            artifact_path=artifact_path,
            executor=executor_for_multipart_upload,
            multipart_info=multipart_info,
            progress_bar=progress_bar,
        )

    def _upload(
        self,
        files_for_normal_upload: Sequence[Tuple[str, str, _FileMultiPartInfo]],
        files_for_multipart_upload: Sequence[Tuple[str, str, _FileMultiPartInfo]],
        progress: Optional[bool] = None,
    ):
        abort_event = Event()
        show_progress = _can_display_progress(progress)
        with Progress(
            "[progress.description]{task.description}",
            BarColumn(bar_width=None),
            "[progress.percentage]{task.percentage:>3.0f}%",
            DownloadColumn(),
            TransferSpeedColumn(),
            TimeRemainingColumn(),
            TimeElapsedColumn(),
            refresh_per_second=1,
            disable=not show_progress,
            expand=True,
        ) as progress_bar, ThreadPoolExecutor(
            max_workers=ENV_VARS.TFY_ARTIFACTS_UPLOAD_MAX_WORKERS
        ) as executor:
            futures: List[Future] = []
            # Note: While this batching is beneficial when there is a large number of files, there is also
            # a rare case risk of the signed url expiring before a request is made to it
            _batch_size = _GENERATE_SIGNED_URL_BATCH_SIZE
            for start_idx in range(0, len(files_for_normal_upload), _batch_size):
                end_idx = min(start_idx + _batch_size, len(files_for_normal_upload))
                if _any_future_has_failed(futures):
                    break
                logger.debug("Generating write signed urls for a batch ...")
                remote_file_paths = [
                    files_for_normal_upload[idx][0] for idx in range(start_idx, end_idx)
                ]
                signed_urls = self.get_signed_urls_for_write(
                    artifact_identifier=self.artifact_identifier,
                    paths=remote_file_paths,
                )
                for idx, signed_url in zip(range(start_idx, end_idx), signed_urls):
                    (
                        upload_path,
                        local_file,
                        multipart_info,
                    ) = files_for_normal_upload[idx]
                    future = executor.submit(
                        self._upload_file,
                        local_file=local_file,
                        artifact_path=upload_path,
                        multipart_info=multipart_info,
                        signed_url=signed_url,
                        abort_event=abort_event,
                        executor_for_multipart_upload=None,
                        progress_bar=progress_bar,
                    )
                    futures.append(future)

            done, not_done = wait(futures, return_when=FIRST_EXCEPTION)
            if len(not_done) > 0:
                abort_event.set()
            for future in not_done:
                future.cancel()
            for future in done:
                if future.exception() is not None:
                    raise future.exception()

            for (
                upload_path,
                local_file,
                multipart_info,
            ) in files_for_multipart_upload:
                self._upload_file(
                    local_file=local_file,
                    artifact_path=upload_path,
                    signed_url=None,
                    multipart_info=multipart_info,
                    executor_for_multipart_upload=executor,
                    progress_bar=progress_bar,
                )

    def _add_file_for_upload(
        self,
        local_file: str,
        artifact_path: str,
        files_for_normal_upload: List[Tuple[str, str, _FileMultiPartInfo]],
        files_for_multipart_upload: List[Tuple[str, str, _FileMultiPartInfo]],
    ):
        local_file = os.path.realpath(local_file)
        if os.path.isdir(local_file):
            raise MlFoundryException(
                "Cannot log a directory as an artifact. Use `log_artifacts` instead"
            )
        upload_path = artifact_path
        upload_path = upload_path.lstrip(posixpath.sep)
        multipart_info = decide_file_parts(local_file)
        if multipart_info.num_parts == 1:
            files_for_normal_upload.append((upload_path, local_file, multipart_info))
        else:
            files_for_multipart_upload.append((upload_path, local_file, multipart_info))

    def _add_dir_for_upload(
        self,
        local_dir: str,
        artifact_path: str,
        files_for_normal_upload: List[Tuple[str, str, _FileMultiPartInfo]],
        files_for_multipart_upload: List[Tuple[str, str, _FileMultiPartInfo]],
    ):
        local_dir = os.path.realpath(local_dir)
        if not os.path.isdir(local_dir):
            raise MlFoundryException(
                "Cannot log a file as a directory. Use `log_artifact` instead"
            )
        dest_path = artifact_path
        dest_path = dest_path.lstrip(posixpath.sep)

        for root, _, file_names in os.walk(local_dir):
            upload_path = dest_path
            if root != local_dir:
                rel_path = os.path.relpath(root, local_dir)
                rel_path = relative_path_to_artifact_path(rel_path)
                upload_path = posixpath.join(dest_path, rel_path)
            for file_name in file_names:
                local_file = os.path.join(root, file_name)
                self._add_file_for_upload(
                    local_file=local_file,
                    artifact_path=upload_path,
                    files_for_normal_upload=files_for_normal_upload,
                    files_for_multipart_upload=files_for_multipart_upload,
                )

    def log_artifacts(
        self,
        src_dest_pairs: Sequence[Tuple[str, Optional[str]]],
        progress: Optional[bool] = None,
    ):
        files_for_normal_upload: List[Tuple[str, str, _FileMultiPartInfo]] = []
        files_for_multipart_upload: List[Tuple[str, str, _FileMultiPartInfo]] = []
        for src, dest in src_dest_pairs:
            if os.path.isdir(src):
                self._add_dir_for_upload(
                    local_dir=src,
                    artifact_path=dest or "",
                    files_for_normal_upload=files_for_normal_upload,
                    files_for_multipart_upload=files_for_multipart_upload,
                )
            else:
                self._add_file_for_upload(
                    local_file=src,
                    artifact_path=dest or "",
                    files_for_normal_upload=files_for_normal_upload,
                    files_for_multipart_upload=files_for_multipart_upload,
                )
        self._upload(
            files_for_normal_upload=files_for_normal_upload,
            files_for_multipart_upload=files_for_multipart_upload,
            progress=progress,
        )

    def list_artifacts(
        self, path=None, page_size=_LIST_FILES_PAGE_SIZE, **kwargs
    ) -> Iterator[FileInfo]:
        if self.artifact_identifier.dataset_id:
            for file_info in client.data_directories.list_files(
                id=str(self.artifact_identifier.dataset_id),
                path=path,
                limit=page_size,
            ):
                yield file_info
        else:
            for file_info in client.artifact_versions.list_files(
                id=str(self.artifact_identifier.artifact_version_id),
                path=path,
                limit=page_size,
            ):
                yield file_info

    def _is_directory(self, artifact_path):
        # TODO: Ideally server should return a flag to indicate if it is a directory
        # For now we can rely on list_artifacts because generally in cloud buckets
        # directories are not stored as separate entities - empty directories ideally do not exist
        for _ in self.list_artifacts(artifact_path, page_size=3):
            return True
        return False

    def _create_download_destination(
        self, src_artifact_path: str, dst_local_dir_path: str
    ) -> str:
        """
        Creates a local filesystem location to be used as a destination for downloading the artifact
        specified by `src_artifact_path`. The destination location is a subdirectory of the
        specified `dst_local_dir_path`, which is determined according to the structure of
        `src_artifact_path`. For example, if `src_artifact_path` is `dir1/file1.txt`, then the
        resulting destination path is `<dst_local_dir_path>/dir1/file1.txt`. Local directories are
        created for the resulting destination location if they do not exist.

        :param src_artifact_path: A relative, POSIX-style path referring to an artifact stored
                                  within the repository's artifact root location.
                                  `src_artifact_path` should be specified relative to the
                                  repository's artifact root location.
        :param dst_local_dir_path: The absolute path to a local filesystem directory in which the
                                   local destination path will be contained. The local destination
                                   path may be contained in a subdirectory of `dst_root_dir` if
                                   `src_artifact_path` contains subdirectories.
        :return: The absolute path to a local filesystem location to be used as a destination
                 for downloading the artifact specified by `src_artifact_path`.
        """
        src_artifact_path = src_artifact_path.rstrip(
            "/"
        )  # Ensure correct dirname for trailing '/'
        dirpath = posixpath.dirname(src_artifact_path)
        local_dir_path = os.path.join(dst_local_dir_path, dirpath)
        local_file_path = os.path.join(dst_local_dir_path, src_artifact_path)
        if not os.path.exists(local_dir_path):
            os.makedirs(local_dir_path, exist_ok=True)
        return local_file_path

    # noinspection PyMethodOverriding
    def _download_file(
        self,
        remote_file_path: str,
        local_path: str,
        progress_bar: Optional[Progress],
        signed_url: Optional[SignedUrl],
        abort_event: Optional[Event] = None,
    ):
        if not remote_file_path:
            raise MlFoundryException(
                f"remote_file_path cannot be None or empty str {remote_file_path}"
            )
        if not signed_url:
            signed_url = self.get_signed_urls_for_read(
                artifact_identifier=self.artifact_identifier, paths=[remote_file_path]
            )[0]

        if progress_bar is None or progress_bar.disable:
            logger.info("Downloading %s to %s", remote_file_path, local_path)

        if progress_bar is not None:
            task_id = progress_bar.add_task(
                f"[green]⬇ {truncate_path_for_progress(remote_file_path, 64)}",
                start=True,
                visible=True,
            )

        def callback(chunk_size: int, total_file_size: int):
            nonlocal task_id
            if progress_bar is not None:
                progress_bar.update(
                    task_id,
                    advance=chunk_size,
                    total=total_file_size,
                )
            if abort_event and abort_event.is_set():
                raise Exception("aborting download")

        _download_file_using_http_uri(
            http_uri=signed_url.signed_url,
            download_path=local_path,
            callback=callback,
        )

        if progress_bar is not None:
            progress_bar.refresh()
        logger.debug("Downloaded %s to %s", remote_file_path, local_path)

    def _download_artifact(
        self,
        src_artifact_path: str,
        dst_local_dir_path: str,
        signed_url: Optional[SignedUrl],
        progress_bar: Optional[Progress] = None,
        abort_event=None,
    ) -> str:
        """
        Download the file artifact specified by `src_artifact_path` to the local filesystem
        directory specified by `dst_local_dir_path`.
        :param src_artifact_path: A relative, POSIX-style path referring to a file artifact
                                    stored within the repository's artifact root location.
                                    `src_artifact_path` should be specified relative to the
                                    repository's artifact root location.
        :param dst_local_dir_path: Absolute path of the local filesystem destination directory
                                    to which to download the specified artifact. The downloaded
                                    artifact may be written to a subdirectory of
                                    `dst_local_dir_path` if `src_artifact_path` contains
                                    subdirectories.
        :param progress_bar: An instance of a Rich progress bar used to visually display the
                              progress of the file download.
        :return: A local filesystem path referring to the downloaded file.
        """
        local_destination_file_path = self._create_download_destination(
            src_artifact_path=src_artifact_path, dst_local_dir_path=dst_local_dir_path
        )
        self._download_file(
            remote_file_path=src_artifact_path,
            local_path=local_destination_file_path,
            signed_url=signed_url,
            abort_event=abort_event,
            progress_bar=progress_bar,
        )
        return local_destination_file_path

    def _get_file_paths_recur(self, src_artifact_dir_path, dst_local_dir_path):
        local_dir = os.path.join(dst_local_dir_path, src_artifact_dir_path)
        # prevent infinite loop, sometimes the dir is recursively included
        # TODO (chiragjn): Check why and when this happens
        dir_content = [
            file_info
            for file_info in self.list_artifacts(src_artifact_dir_path)
            if file_info.path != "." and file_info.path != src_artifact_dir_path
        ]
        if not dir_content:  # empty dir
            if not os.path.exists(local_dir):
                os.makedirs(local_dir, exist_ok=True)
        else:
            for file_info in dir_content:
                if file_info.is_dir:
                    yield from self._get_file_paths_recur(
                        src_artifact_dir_path=file_info.path,
                        dst_local_dir_path=dst_local_dir_path,
                    )
                else:
                    yield file_info.path, dst_local_dir_path

    def download_artifacts(  # noqa: C901
        self,
        artifact_path: str,
        dst_path: Optional[str] = None,
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> str:
        """
        Download an artifact file or directory to a local directory if applicable, and return a
        local path for it. The caller is responsible for managing the lifecycle of the downloaded artifacts.

        Args:
            artifact_path: Relative source path to the desired artifacts.
            dst_path: Absolute path of the local filesystem destination directory to which to
                download the specified artifacts. This directory must already exist.
                If unspecified, the artifacts will either be downloaded to a new
                uniquely-named directory.
            overwrite: if to overwrite the files at/inside `dst_path` if they exist
            progress: Show or hide progress bar

        Returns:
            str: Absolute path of the local filesystem location containing the desired artifacts.
        """

        show_progress = _can_display_progress(user_choice=progress)

        is_dir_temp = False
        if dst_path is None:
            dst_path = tempfile.mkdtemp()
            is_dir_temp = True

        dst_path = os.path.abspath(dst_path)
        if is_dir_temp:
            logger.info(
                f"Using temporary directory {dst_path} as the download directory"
            )

        if not os.path.exists(dst_path):
            raise MlFoundryException(
                message=(
                    "The destination path for downloaded artifacts does not"
                    " exist! Destination path: {dst_path}".format(dst_path=dst_path)
                ),
            )
        elif not os.path.isdir(dst_path):
            raise MlFoundryException(
                message=(
                    "The destination path for downloaded artifacts must be a directory!"
                    " Destination path: {dst_path}".format(dst_path=dst_path)
                ),
            )

        progress_bar = Progress(
            "[progress.description]{task.description}",
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            DownloadColumn(),
            TransferSpeedColumn(),
            TimeRemainingColumn(),
            TimeElapsedColumn(),
            refresh_per_second=1,
            disable=not show_progress,
            expand=True,
        )

        try:
            progress_bar.start()
            # empty string is treated as root directory
            if not artifact_path or self._is_directory(artifact_path):
                futures: List[Future] = []
                file_paths: List[Tuple[str, str]] = []
                abort_event = Event()

                # Check if any file is being overwritten before downloading them
                for file_path, download_dest_path in self._get_file_paths_recur(
                    src_artifact_dir_path=artifact_path, dst_local_dir_path=dst_path
                ):
                    final_file_path = os.path.join(download_dest_path, file_path)

                    # There would be no overwrite if temp directory is being used
                    if (
                        not is_dir_temp
                        and os.path.exists(final_file_path)
                        and not overwrite
                    ):
                        raise MlFoundryException(
                            f"File already exists at {final_file_path}, aborting download "
                            f"(set `overwrite` flag to overwrite this and any subsequent files)"
                        )
                    file_paths.append((file_path, download_dest_path))

                with ThreadPoolExecutor(
                    max_workers=ENV_VARS.TFY_ARTIFACTS_DOWNLOAD_MAX_WORKERS
                ) as executor:
                    # Note: While this batching is beneficial when there is a large number of files, there is also
                    # a rare case risk of the signed url expiring before a request is made to it
                    batch_size = _GENERATE_SIGNED_URL_BATCH_SIZE
                    for start_idx in range(0, len(file_paths), batch_size):
                        end_idx = min(start_idx + batch_size, len(file_paths))
                        if _any_future_has_failed(futures):
                            break
                        logger.debug("Generating read signed urls for a batch ...")
                        remote_file_paths = [
                            file_paths[idx][0] for idx in range(start_idx, end_idx)
                        ]
                        signed_urls = self.get_signed_urls_for_read(
                            artifact_identifier=self.artifact_identifier,
                            paths=remote_file_paths,
                        )
                        for idx, signed_url in zip(
                            range(start_idx, end_idx), signed_urls
                        ):
                            file_path, download_dest_path = file_paths[idx]
                            future = executor.submit(
                                self._download_artifact,
                                src_artifact_path=file_path,
                                dst_local_dir_path=download_dest_path,
                                signed_url=signed_url,
                                abort_event=abort_event,
                                progress_bar=progress_bar,
                            )
                            futures.append(future)

                    done, not_done = wait(futures, return_when=FIRST_EXCEPTION)
                    if len(not_done) > 0:
                        abort_event.set()
                    for future in not_done:
                        future.cancel()
                    for future in done:
                        if future.exception() is not None:
                            raise future.exception()

                    output_dir = os.path.join(dst_path, artifact_path)
                    return output_dir
            else:
                return self._download_artifact(
                    src_artifact_path=artifact_path,
                    dst_local_dir_path=dst_path,
                    signed_url=None,
                    progress_bar=progress_bar,
                )
        except Exception as err:
            if is_dir_temp:
                logger.info(
                    f"Error encountered, removing temporary download directory at {dst_path}"
                )
                rmtree(dst_path)  # remove temp directory alongside it's contents
            raise err

        finally:
            progress_bar.stop()



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.ml._autogen.client.models.subject import Subject
from truefoundry.pydantic_v1 import BaseModel, Field, StrictInt, StrictStr


class RunInfoDto(BaseModel):
    """
    RunInfoDto
    """

    run_uuid: StrictStr = Field(...)
    run_id: StrictStr = Field(...)
    experiment_id: StrictStr = Field(...)
    name: StrictStr = Field(...)
    fqn: StrictStr = Field(...)
    user_id: Optional[StrictStr] = None
    created_by_subject: Subject = Field(...)
    status: StrictStr = Field(...)
    start_time: Optional[StrictInt] = None
    end_time: Optional[StrictInt] = None
    artifact_uri: Optional[StrictStr] = None
    lifecycle_stage: Optional[StrictStr] = None
    description: Optional[StrictStr] = None
    __properties = [
        "run_uuid",
        "run_id",
        "experiment_id",
        "name",
        "fqn",
        "user_id",
        "created_by_subject",
        "status",
        "start_time",
        "end_time",
        "artifact_uri",
        "lifecycle_stage",
        "description",
    ]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> RunInfoDto:
        """Create an instance of RunInfoDto from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of created_by_subject
        if self.created_by_subject:
            _dict["created_by_subject"] = self.created_by_subject.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> RunInfoDto:
        """Create an instance of RunInfoDto from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return RunInfoDto.parse_obj(obj)

        _obj = RunInfoDto.parse_obj(
            {
                "run_uuid": obj.get("run_uuid"),
                "run_id": obj.get("run_id"),
                "experiment_id": obj.get("experiment_id"),
                "name": obj.get("name"),
                "fqn": obj.get("fqn"),
                "user_id": obj.get("user_id"),
                "created_by_subject": Subject.from_dict(obj.get("created_by_subject"))
                if obj.get("created_by_subject") is not None
                else None,
                "status": obj.get("status"),
                "start_time": obj.get("start_time"),
                "end_time": obj.get("end_time"),
                "artifact_uri": obj.get("artifact_uri"),
                "lifecycle_stage": obj.get("lifecycle_stage"),
                "description": obj.get("description"),
            }
        )
        return _obj



from __future__ import annotations

import os
from typing import Any, Generator, List, Optional

from truefoundry.autodeploy.constants import MAX_FILE_SIZE_READ
from truefoundry.autodeploy.tools.base import (
    Event,
    Message,
    RequestEvent,
    ResponseEvent,
    Rule,
    Tool,
)
from truefoundry.pydantic_v1 import BaseModel, Field


class Line(BaseModel):
    line_number: int
    content: str


class ReadFile(Tool):
    description = """
Read contents of a file.
Avoid reading *.lock type files as they tend to be large
"""

    class Request(RequestEvent):
        path: str = Field(
            ...,
            pattern=r"^[a-zA-Z0-9\.]{1}.*$",
            description="File path to open. ",
        )

    class Response(ResponseEvent):
        data: Optional[List[Line]] = Field(
            None,
            description="Content of the file.",
        )
        error: Optional[str] = Field(
            None,
            description="Error while opening a file.",
        )

    def __init__(
        self,
        project_root_path: str,
    ):
        self.project_root_path = project_root_path

    def run(self, request: ReadFile.Request) -> Generator[Event, Any, ResponseEvent]:
        yield Rule(message="[bold green] Read files[/]")
        yield Message(
            message=f"[bold cyan]Processing:[/] Reading file at [magenta]{request.path}[/] and extracting details..."
        )

        try:
            file_path = os.path.join(self.project_root_path, request.path)
            file_size = os.path.getsize(file_path)

            if file_size > MAX_FILE_SIZE_READ:
                return ReadFile.Response(
                    error=f"File size exceeds {MAX_FILE_SIZE_READ} bytes limit"
                )

            with open(file_path, "r", encoding="utf8") as f:
                response = ReadFile.Response(data=[])
                for i, line in enumerate(f):
                    response.data.append(Line(line_number=i + 1, content=line))
                return response
        except FileNotFoundError as ex:
            return ReadFile.Response(error=str(ex))
        except UnicodeDecodeError:
            return ReadFile.Response(
                error="This is a binary file, reading it is prohibited."
            )



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    StrictStr,
    conlist,
    constr,
    validator,
)


class ModelVersionEnvironment(BaseModel):
    """
    Runtime environment describing python version and dependencies for running the model  # noqa: E501
    """

    python_version: Optional[constr(strict=True)] = Field(
        default=None, description="Python version to use when running the model"
    )
    pip_packages: Optional[conlist(StrictStr)] = Field(
        default=None, description="pip dependencies needed for running the model"
    )
    __properties = ["python_version", "pip_packages"]

    @validator("python_version")
    def python_version_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if value is None:
            return value

        if not re.match(r"^\d+(\.\d+){1,2}([\-\.a-z0-9]+)?$", value):
            raise ValueError(
                r"must validate the regular expression /^\d+(\.\d+){1,2}([\-\.a-z0-9]+)?$/"
            )
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ModelVersionEnvironment:
        """Create an instance of ModelVersionEnvironment from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ModelVersionEnvironment:
        """Create an instance of ModelVersionEnvironment from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ModelVersionEnvironment.parse_obj(obj)

        _obj = ModelVersionEnvironment.parse_obj(
            {
                "python_version": obj.get("python_version"),
                "pip_packages": obj.get("pip_packages"),
            }
        )
        return _obj



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, Dict, Optional

from truefoundry.ml._autogen.client.models.source import Source
from truefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    StrictStr,
    conint,
    constr,
    validator,
)


class ArtifactManifest(BaseModel):
    """
    Artifact Version manifest.  # noqa: E501
    """

    name: Optional[constr(strict=True, max_length=256)] = Field(
        default=None, description="Name of the entity"
    )
    description: Optional[constr(strict=True, max_length=512)] = None
    metadata: Dict[str, Any] = Field(
        default=...,
        description='Key value metadata. Should be valid JSON. For e.g. `{"business-unit": "sales", "quality": "good", "rating": 4.5}`',
    )
    version_alias: Optional[constr(strict=True, max_length=128)] = Field(
        default=None,
        description="Version alias is alternate, ideally human readable, version string to reference an artifact version. It should start with `v` followed by alphanumeric and it can include `.` and `-` in between (e.g. `v1.0.0`, `v1-prod`, `v3-dev`, etc)",
    )
    ml_repo: Optional[constr(strict=True)] = Field(
        default=None, description="Name of the ML Repo"
    )
    version: Optional[conint(strict=True, ge=1)] = Field(
        default=None, description="Version of the entity"
    )
    type: Optional[StrictStr] = "artifact-version"
    source: Source = Field(...)
    step: Optional[conint(strict=True, ge=0)] = Field(
        default=0,
        description="Step/Epoch number in an iterative training loop the artifact version was created. Generally useful when logging a model version from a MLRepo Run",
    )
    run_id: Optional[StrictStr] = Field(
        default=None,
        description="ID of the MLRepo Run that generated the artifact version",
    )
    __properties = [
        "name",
        "description",
        "metadata",
        "version_alias",
        "ml_repo",
        "version",
        "type",
        "source",
        "step",
        "run_id",
    ]

    @validator("name")
    def name_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if value is None:
            return value

        if not re.match(r"^[A-Za-z0-9_\-]+$", value):
            raise ValueError(
                r"must validate the regular expression /^[A-Za-z0-9_\-]+$/"
            )
        return value

    @validator("version_alias")
    def version_alias_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if value is None:
            return value

        if not re.match(r"^v[a-zA-Z0-9.-]*([a-zA-Z0-9]+)$", value):
            raise ValueError(
                r"must validate the regular expression /^v[a-zA-Z0-9.-]*([a-zA-Z0-9]+)$/"
            )
        return value

    @validator("ml_repo")
    def ml_repo_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if value is None:
            return value

        if not re.match(r"^[a-zA-Z][a-zA-Z0-9\-]{1,98}[a-zA-Z0-9]$", value):
            raise ValueError(
                r"must validate the regular expression /^[a-zA-Z][a-zA-Z0-9\-]{1,98}[a-zA-Z0-9]$/"
            )
        return value

    @validator("type")
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in ("artifact-version",):
            raise ValueError("must be one of enum values ('artifact-version')")
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ArtifactManifest:
        """Create an instance of ArtifactManifest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of source
        if self.source:
            _dict["source"] = self.source.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> ArtifactManifest:
        """Create an instance of ArtifactManifest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return ArtifactManifest.parse_obj(obj)

        _obj = ArtifactManifest.parse_obj(
            {
                "name": obj.get("name"),
                "description": obj.get("description"),
                "metadata": obj.get("metadata"),
                "version_alias": obj.get("version_alias"),
                "ml_repo": obj.get("ml_repo"),
                "version": obj.get("version"),
                "type": obj.get("type")
                if obj.get("type") is not None
                else "artifact-version",
                "source": Source.from_dict(obj.get("source"))
                if obj.get("source") is not None
                else None,
                "step": obj.get("step") if obj.get("step") is not None else 0,
                "run_id": obj.get("run_id"),
            }
        )
        return _obj



BUILDKIT_SECRET_MOUNT_PIP_CONF_ID = "pip.conf"
PIP_CONF_BUILDKIT_SECRET_MOUNT = (
    f"--mount=type=secret,id={BUILDKIT_SECRET_MOUNT_PIP_CONF_ID}"
)
PIP_CONF_SECRET_MOUNT_AS_ENV = (
    f"PIP_CONFIG_FILE=/run/secrets/{BUILDKIT_SECRET_MOUNT_PIP_CONF_ID}"
)

BUILDKIT_SECRET_MOUNT_UV_CONF_ID = "uv.toml"
UV_CONF_BUILDKIT_SECRET_MOUNT = (
    f"--mount=type=secret,id={BUILDKIT_SECRET_MOUNT_UV_CONF_ID}"
)
UV_CONF_SECRET_MOUNT_AS_ENV = (
    f"UV_CONFIG_FILE=/run/secrets/{BUILDKIT_SECRET_MOUNT_UV_CONF_ID}"
)

BUILDKIT_SECRET_MOUNT_UV_ENV_ID = "uv.env"
UV_ENV_BUILDKIT_SECRET_MOUNT = (
    f"--mount=type=secret,id={BUILDKIT_SECRET_MOUNT_UV_ENV_ID}"
)
UV_ENV_SECRET_MOUNT_AS_ENV = f". /run/secrets/{BUILDKIT_SECRET_MOUNT_UV_ENV_ID}"

BUILDKIT_SECRET_MOUNT_POETRY_ENV_ID = "poetry.env"
POETRY_ENV_BUILDKIT_SECRET_MOUNT = (
    f"--mount=type=secret,id={BUILDKIT_SECRET_MOUNT_POETRY_ENV_ID}"
)
POETRY_ENV_SECRET_MOUNT_AS_ENV = f". /run/secrets/{BUILDKIT_SECRET_MOUNT_POETRY_ENV_ID}"



# This script is used as the main application file for spark applications
# when the application to be run is a notebook, the actual notebook to be
# executed is passed as an argument to this script.


import argparse
import os
import sys

import boto3
import nbformat
import papermill as pm
from botocore.client import Config
from nbconvert import HTMLExporter


def convert_notebook_to_html(notebook_path, output_html_path):
    """
    Convert a Jupyter notebook to an HTML file.

    Args:
        notebook_path: Path to the input notebook (.ipynb)
        output_html_path: Path for the output HTML file (.html)
    """
    print(f"Converting notebook {notebook_path} to HTML...")
    try:
        with open(notebook_path, "r", encoding="utf-8") as f:
            notebook_content = nbformat.read(f, as_version=4)

        html_exporter = HTMLExporter()
        # Use lab for https://nbconvert.readthedocs.io/en/latest/customizing.html#where-are-nbconvert-templates-installed
        html_exporter.template_name = "lab"
        (body, _) = html_exporter.from_notebook_node(notebook_content)

        with open(output_html_path, "w", encoding="utf-8") as f:
            f.write(body)
        print(f"Successfully converted notebook to HTML: {output_html_path}")
    except Exception as e:
        print(f"Error converting notebook to HTML: {e}")
        raise


def upload_file_to_s3(file_path, bucket_name, s3_key):
    print(f"Uploading {file_path} to s3://{bucket_name}/{s3_key}...")
    # Use s3proxy for pushing data to s3
    # The JWT token is already available in the pod
    aws_access_key_id = os.environ.get("SPARK_APPLICATION_EVENT_LOG_JWT_TOKEN")
    aws_secret_access_key = os.environ.get("TFY_NOTEBOOK_OUTPUT_S3_SECRET_KEY")
    s3_endpoint_url = os.environ.get("S3_PROXY_URL")

    # Needed for the issue https://github.com/gaul/s3proxy/issues/765
    s3_config = Config(
        request_checksum_calculation="when_required",
        response_checksum_validation="when_required",
    )
    try:
        client = boto3.client(
            "s3",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            endpoint_url=s3_endpoint_url,
            config=s3_config,
        )

        with open(file_path, "rb") as data:
            client.put_object(Bucket=bucket_name, Key=s3_key, Body=data)
        print(f"Successfully uploaded {file_path} to s3://{bucket_name}/{s3_key}")
    except Exception as e:
        print(f"Error uploading file to S3: {e}")
        raise


def execute_notebook(notebook_path, output_path="/tmp/output.ipynb", parameters=None):
    """
    Execute a Jupyter notebook using papermill.

    Args:
        notebook_path: Path to the input notebook
        output_path: Path for the output notebook
        parameters: Dictionary of parameters to pass to the notebook

    Raises:
        Exception: If notebook execution fails
    """
    if parameters is None:
        parameters = {}

    print(f"Starting execution of notebook: {notebook_path}")
    notebook_type = os.environ.get("TFY_NOTEBOOK_TYPE", "").lower()
    kernel_mapping = {"python": "python3", "scala": "scala"}

    if notebook_type not in kernel_mapping:
        supported_types = ", ".join(kernel_mapping.keys())
        raise ValueError(
            f"Unsupported notebook type: '{notebook_type}'. "
            f"Supported types: [{supported_types}]"
        )

    kernel_name = kernel_mapping[notebook_type]

    pm.execute_notebook(
        input_path=notebook_path,
        output_path=output_path,
        parameters=parameters,
        # TODO(gw): Replace with kernel name for venv
        kernel_name=kernel_name,
        # Log cell by cell execution output
        # TODO(gw): Output logs to a file instead, so that they aren't merged with the container's logs
        log_output=True,
        stdout_file=sys.stdout,
        stderr_file=sys.stderr,
        cwd=os.environ.get("TFY_WORKDIR"),
    )
    print(f"Successfully executed notebook: {notebook_path}")


def validate_env_vars():
    keys = [
        "TFY_NOTEBOOK_OUTPUT_S3_KEY",
        "TFY_NOTEBOOK_OUTPUT_S3_BUCKET",
        "SPARK_APPLICATION_EVENT_LOG_JWT_TOKEN",
        "TFY_NOTEBOOK_OUTPUT_S3_SECRET_KEY",
        "TFY_NOTEBOOK_TYPE",
        "TFY_WORKDIR",
    ]
    unset_keys = [key for key in keys if not os.environ.get(key)]
    if unset_keys:
        raise ValueError(
            f"Environment variables {unset_keys} are not set."
            f"Contact you tenant-admin to configure storage bucket on the control plane "
            f"to enable uploading spark notebook outputs."
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Execute a Jupyter notebook using papermill for Spark applications"
    )
    parser.add_argument("notebook_path", help="Path to the notebook file to execute")
    args = parser.parse_args()

    # Since failure to upload is considered a job failure, fail the job even before it run if uploads cannot happen
    validate_env_vars()

    output_notebook_path = "/tmp/output.ipynb"

    # This would be the same as the default bucket used by servicefoundry-server
    s3_bucket = os.environ.get("TFY_NOTEBOOK_OUTPUT_S3_BUCKET")
    # This would be something like sparkjob-events/<tenant-id>/output-notebooks/<application-id>/<jobrun-name>/output.html
    s3_key = os.environ.get("TFY_NOTEBOOK_OUTPUT_S3_KEY")

    try:
        execute_notebook(args.notebook_path, output_path=output_notebook_path)

        # The following may also be modeled as an entrypoint
        # https://papermill.readthedocs.io/en/latest/extending-entry-points.html
        # Will take that up with next iteration where we save the executed notebook periodically
        print("Converting notebook to HTML and uploading to S3...")
        html_output_path = "/tmp/output.html"
        convert_notebook_to_html(
            notebook_path=output_notebook_path, output_html_path=html_output_path
        )
        upload_file_to_s3(
            file_path=html_output_path, bucket_name=s3_bucket, s3_key=s3_key
        )
        print(f"Successfully uploaded HTML to s3://{s3_bucket}/{s3_key}")

    except Exception as e:
        print(f"Error executing notebook {args.notebook_path}: {e}")
        print(
            "Exiting with status code 1 to signal failure to parent process/orchestrator"
        )
        sys.exit(1)



import json
import os
import posixpath
import shutil
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    ArtifactType,
)
from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.log_types.artifacts.artifact import (
    ArtifactVersionInternalMetadata,
    _log_artifact_version_helper,
)
from truefoundry.ml.log_types.artifacts.constants import (
    FILES_DIR,
    INTERNAL_METADATA_PATH,
)
from truefoundry.ml.log_types.artifacts.utils import _make_dest_to_src_map_from_dir
from truefoundry.ml.log_types.image.constants import (
    DEFAULT_IMAGE_FORMAT,
    IMAGE_KEY_REGEX,
    IMAGE_METADATA_FILE_NAME,
    MISSING_PILLOW_PACKAGE_MESSAGE,
)
from truefoundry.ml.log_types.image.image_normalizer import normalize_image
from truefoundry.ml.log_types.image.types import BoundingBoxGroups, ClassGroups
from truefoundry.ml.log_types.pydantic_base import PydanticBase
from truefoundry.ml.run_utils import get_module

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    import numpy  # noqa: F401
    import PIL.Image
    from numpy.typing import NDArray  # noqa: F401

    from truefoundry.ml.mlfoundry_run import MlFoundryRun

ImageLikeNDArray = Union[
    "NDArray[numpy.int8]",
    "NDArray[numpy.uint8]",
    "NDArray[numpy.float32]",
    "NDArray[numpy.float64]",
    "NDArray[numpy.bool_]",
]
DataOrPathType = Union[str, Path, ImageLikeNDArray, "PIL.Image.Image"]
ClassGroupsType = Dict[str, Union[str, List[str]]]
BBoxGrouptype = Dict[str, List[Dict[str, Any]]]


def validate_key_name(key: str):
    if not key or not IMAGE_KEY_REGEX.match(key):
        raise MlFoundryException(
            f"Invalid run image key: {key} should only contain alphanumeric, hyphen or underscore"
        )


class ImageVersionInternalMetadata(ArtifactVersionInternalMetadata):
    image_file: str
    image_metadata_file: str


class Image:
    def __init__(
        self,
        data_or_path: DataOrPathType,
        caption: Optional[str] = None,
        class_groups: Optional[ClassGroupsType] = None,
        bbox_groups: Optional[BBoxGrouptype] = None,
    ):
        """Represent and log image using this class in `mlfoundry`.

        You can initialize `truefoundry.ml.Image` by either by using a local path
        or you can use a numpy array / PIL.Image object.

        If you are using numpy array, we only support the following data types,
        - bool
        - integer [0 - 255]
        - unsigned integer [0 - 255]
        - float [0.0 - 1.0]

        Any out of range value will be clipped.

        As for array shape/dim, we follow the following structures,
        - H x W (Grayscale)
        - H x W x 1 (Grayscale)
        - H x W x 3 (an RGB channel order is assumed)
        - H x W x 4 (an RGBA channel order is assumed)

        `PIL` package is required to log images. To install the `PIL` package,
        run `pip install pillow`.

        We can also log class names and bounding boxes associated with the image.
        Class names and bounding boxes should be always grouped under `actuals` or
        `predictions`. For example, if we have an image where the ground truth class
        is "cat" and predicted class is "dog", we can represent it like,

        ```python
        from truefoundry.ml import Image

        Image(
            data_or_path=imarray,
            class_groups={"actuals": "dog", "predictions": "cat"}
        )
        ```

        You can define a bounding box using the following dictionary structure,
        ```python
        {
            "position": {"min_x": 15, "min_y": 5, "max_x": 20, "max_y": 30}, # required, defines the position of the bounding box
                                                                             # (min_x, min_y) defines the top left and
                                                                             # (max_x, max_y) defines the bottom right corner of the box.
            "class_name": "dog", # required, the class name of the bounding box
            "caption": "dog", # optional, the caption of the bounding box.
                              # If not passed, the class name is set as caption.
        }
        ```

        Args:
            data_or_path (Union[str, Path, "numpy.ndarray", "PIL.Image.Image"]):
                Either the local path or the image object (Numpy array or PIL Image).
            caption (Optional[str], optional): A string caption or label for the image.
            class_groups (Optional[Dict[str, Union[str, List[str]]]], optional):
                Class names associated with the image. Expects class name(s) grouped by
                `predictions` or `actuals`.
            bbox_groups (Optional[Dict[str, List[Dict]]], optional): Bounding boxes
                associated with the image. Expects bounding boxes grouped by `predictions`
                or `actuals`.

        Examples:
        ### Logging images with caption and class names

        ```python
        from truefoundry.ml import get_client, Image
        import numpy as np

        client = get_client()
        run = client.create_run(
            ml_repo="my-classification-project",
        )

        imarray = np.random.randint(low=0, high=256, size=(100, 100, 3))

        images_to_log = {
            "logged-image-array": Image(
                data_or_path=imarray,
                caption="testing image logging",
                class_groups={"actuals": "dog", "predictions": "cat"},
            ),
        }
        run.log_images(images_to_log, step=1)

        run.end()
        ```

        ### Logging images for a multi-label classification problem

        ```python
        from truefoundry.ml import Image

        images_to_log = {
            "logged-image-array": Image(
                data_or_path=imarray,
                caption="testing image logging",
                class_groups={"actuals": ["dog", "human"], "predictions": ["cat", "human"]},
            ),
        }

        run.log_images(images_to_log, step=1)
        ```

        ### Logging images with bounding boxes

        ```python
        from truefoundry.ml import Image

        images_to_log = {
            "logged-image-array": Image(
                data_or_path=imarray,
                caption="testing image logging",
                bbox_groups={
                    "predictions": [
                        {
                            "position": {"min_x": 5, "min_y": 5, "max_x": 20, "max_y": 30},
                            "class_name": "cat",
                        }
                    ],
                    "actuals": [
                        {
                            "position": {"min_x": 15, "min_y": 5, "max_x": 20, "max_y": 30},
                            "class_name": "dog",
                            "caption": "dog",
                        }
                    ],
                },
            ),
        }

        run.log_images(images_to_log, step=1)
        ```
        """

        self._caption: Optional[str] = None
        self._class_groups: Optional[ClassGroups] = None
        self._bbox_groups: Optional[BoundingBoxGroups] = None

        self._image: Optional["PIL.Image.Image"] = None
        self._image_artifact_path = None
        self._local_image_path: Optional[str] = None

        self._init_image(data_or_path)
        self._init_metadata(
            caption=caption, class_groups=class_groups, bbox_groups=bbox_groups
        )

    @property
    def image(self) -> "PIL.Image.Image":
        if self._image is None:
            raise MlFoundryException("Image is not initialized")
        return self._image

    @property
    def caption(self) -> Optional[str]:
        return self._caption

    @property
    def class_groups(self) -> Optional[ClassGroups]:
        return self._class_groups

    @property
    def bbox_groups(self) -> Optional[BoundingBoxGroups]:
        return self._bbox_groups

    def save_image_to_local_dir(self, local_dir) -> str:
        if self._local_image_path is None:
            image_file_name = self._serialize_and_save_image_as_artifact(local_dir)
        else:
            image_file_name = self._copy_local_image_as_artifact(local_dir)

        return image_file_name

    def _serialize_and_save_image_as_artifact(self, local_dir: str) -> str:
        file_name = f"image.{DEFAULT_IMAGE_FORMAT}"
        local_path = os.path.join(local_dir, file_name)
        self.image.save(local_path)

        return file_name

    def _copy_local_image_as_artifact(self, local_dir: str) -> str:
        assert self._local_image_path is not None
        file_name = os.path.basename(self._local_image_path)
        new_local_image_path = os.path.join(local_dir, file_name)
        shutil.copy2(self._local_image_path, new_local_image_path)
        return file_name

    def _init_image(self, data_or_path: DataOrPathType):
        pil_image_module = get_module(
            module_name="PIL.Image",
            required=True,
            error_message=MISSING_PILLOW_PACKAGE_MESSAGE,
        )
        if isinstance(data_or_path, (str, Path)):
            self._local_image_path = os.path.abspath(data_or_path)
            with pil_image_module.open(data_or_path) as image:
                image.load()
                self._image = image
        else:
            self._image = normalize_image(data_or_path)

    def _init_metadata(
        self,
        caption: Optional[str],
        class_groups: Optional[ClassGroupsType],
        bbox_groups: Optional[Dict[str, Any]],
    ):
        if caption is not None:
            self._caption = str(caption)

        if class_groups:
            self._class_groups = ClassGroups.parse_obj(class_groups)

        if bbox_groups:
            self._bbox_groups = BoundingBoxGroups.parse_obj(bbox_groups)

    def _to_dict(self) -> Dict[str, Any]:
        dict_ = {}
        dict_["caption"] = self._caption
        dict_["class_groups"] = (
            self.class_groups.to_dict() if self.class_groups is not None else None
        )
        dict_["bbox_groups"] = (
            self.bbox_groups.to_dict() if self.bbox_groups is not None else None
        )

        return dict_

    def save(
        self,
        run: "MlFoundryRun",
        key: str,
        step: int = 0,
    ):
        validate_key_name(key)

        # creating temp dir, files directory, .truefoundry directory
        temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")

        local_internal_metadata_path = os.path.join(
            temp_dir.name, INTERNAL_METADATA_PATH
        )
        os.makedirs(os.path.dirname(local_internal_metadata_path), exist_ok=True)

        local_files_dir = os.path.join(temp_dir.name, FILES_DIR)
        os.makedirs(local_files_dir, exist_ok=True)

        try:
            # saving the image file
            image_file_name = self.save_image_to_local_dir(local_files_dir)

            # saving the image_metadata
            local_image_metadata_file_path = os.path.join(
                local_files_dir, IMAGE_METADATA_FILE_NAME
            )
            with open(local_image_metadata_file_path, "w") as fp:
                fp.write(ImageRunLogType(value=self._to_dict()).json())

            # creating and saving the internal_metadata file
            internal_metadata = ImageVersionInternalMetadata(
                files_dir=FILES_DIR,
                # joining with posixpath for avoiding backslash in case logged from windows machine
                image_file=posixpath.join(FILES_DIR, image_file_name),
                image_metadata_file=posixpath.join(FILES_DIR, IMAGE_METADATA_FILE_NAME),
            )
            with open(local_internal_metadata_path, "w") as f:
                json.dump(internal_metadata.dict(), f)

        except Exception as e:
            temp_dir.cleanup()
            raise MlFoundryException("Failed to log Image") from e

        _log_artifact_version_helper(
            run=run,
            name=key,
            artifact_type=ArtifactType.IMAGE,
            artifact_dir=temp_dir,
            dest_to_src_map=_make_dest_to_src_map_from_dir(root_dir=temp_dir.name),
            internal_metadata=internal_metadata,
            step=step,
        )


class ImageRunLogType(PydanticBase):
    value: Dict[str, Any]

    @staticmethod
    def get_log_type() -> str:
        return "image"



import logging

from rich.logging import RichHandler

from truefoundry.autodeploy.constants import DEBUG

logger = logging.getLogger("autodeploy")

level = logging.DEBUG if DEBUG else logging.NOTSET
handler = RichHandler(level=level, show_path=False)
handler.setLevel(level)
logger.addHandler(handler)
logger.setLevel(level)



from typing import Any, Dict, List, Optional, Union

from truefoundry.deploy._autogen.models import (
    DockerFileBuild,
    Pip,
    PythonBuild,
    SparkBuild,
    TaskDockerFileBuild,
    TaskPySparkBuild,
    TaskPythonBuild,
)
from truefoundry.deploy.builder.builders import get_builder
from truefoundry.deploy.builder.builders.tfy_notebook_buildpack.dockerfile_template import (
    NotebookImageBuild,
)
from truefoundry.pydantic_v1 import BaseModel, Field


class _BuildConfig(BaseModel):
    __root__: Union[
        DockerFileBuild,
        PythonBuild,
        NotebookImageBuild,
        TaskPythonBuild,
        TaskDockerFileBuild,
        SparkBuild,
        TaskPySparkBuild,
    ] = Field(discriminator="type")


def build(
    build_configuration: Union[BaseModel, Dict[str, Any]],
    tag: str,
    extra_opts: Optional[List[str]] = None,
):
    build_configuration = _BuildConfig.parse_obj(build_configuration).__root__
    if isinstance(build_configuration, TaskPythonBuild):
        build_configuration_dict = build_configuration.dict()
        build_configuration_dict.update({"type": "tfy-python-buildpack", "command": ""})
        build_configuration = PythonBuild.parse_obj(build_configuration_dict)
    if isinstance(build_configuration, TaskDockerFileBuild):
        build_configuration_dict = build_configuration.dict()
        build_configuration_dict.update({"type": "dockerfile"})
        build_configuration = DockerFileBuild.parse_obj(build_configuration_dict)
    builder = get_builder(build_configuration.type)
    return builder(
        build_configuration=build_configuration,
        tag=tag,
        extra_opts=extra_opts,
    )


if __name__ == "__main__":
    # TODO: remove these and write tests for Build class to dockerfile content.
    import os
    from tempfile import TemporaryDirectory

    with TemporaryDirectory() as local_dir:
        dockerfile_path = os.path.join(local_dir, "Dockerfile.test")
        with open(dockerfile_path, "w", encoding="utf8") as fp:
            fp.write("from postgres:latest")

        build_config = DockerFileBuild(
            type="dockerfile",
            build_context_path=local_dir,
            dockerfile_path=dockerfile_path,
        )

        build(tag="docker-test", build_configuration=build_config)

    with TemporaryDirectory() as local_dir:
        requirements_path = os.path.join(local_dir, "requirements.txt")
        with open(requirements_path, "w", encoding="utf8") as fp:
            fp.write("requests")

        python_path = os.path.join(local_dir, "main.py")
        with open(python_path, "w", encoding="utf8") as fp:
            fp.write("import requests; print(requests.__version__);")

        build_config = PythonBuild(
            type="tfy-python-buildpack",
            build_context_path=local_dir,
            command=["python", "main.py"],
            python_version="3.7.13",
        )
        build(tag="python-test-exec", build_configuration=build_config)
        build_config = PythonBuild(
            type="tfy-python-buildpack",
            build_context_path=local_dir,
            command="python main.py",
            python_version="3.8",
        )
        build(tag="python-test-shell", build_configuration=build_config)

    with TemporaryDirectory() as local_dir:
        requirements_path = os.path.join(local_dir, "requirements.txt")
        with open(requirements_path, "w", encoding="utf8") as fp:
            fp.write("numpy")

        python_path = os.path.join(local_dir, "main.py")
        with open(python_path, "w", encoding="utf8") as fp:
            fp.write("import numpy; import requests; print(requests.__version__);")

        build_config = PythonBuild(
            type="tfy-python-buildpack",
            build_context_path=local_dir,
            command=["python", "main.py"],
            python_version="3.7.13",
            pip_packages=["requests"],
        )
        build(tag="python-pip-and-req-file", build_configuration=build_config)

    with TemporaryDirectory() as local_dir:
        python_path = os.path.join(local_dir, "main.py")
        with open(python_path, "w", encoding="utf8") as fp:
            fp.write("import requests; print(requests.__version__);")

        build_config = PythonBuild(
            type="tfy-python-buildpack",
            build_context_path=local_dir,
            command=["python", "main.py"],
            python_version="3.7.13",
            pip_packages=["requests>0.1"],
        )
        build(tag="python-only-pip", build_configuration=build_config)

    with TemporaryDirectory() as local_dir:
        python_path = os.path.join(local_dir, "main.py")
        with open(python_path, "w", encoding="utf8") as fp:
            fp.write("print(1)")

        build_config = PythonBuild(
            type="tfy-python-buildpack",
            build_context_path=local_dir,
            command=["python", "main.py"],
            python_version="3.7.13",
        )
        build(tag="python-only-no-pip-no-req", build_configuration=build_config)



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional, Union

from truefoundry.ml._autogen.client.models.stop import Stop
from truefoundry.pydantic_v1 import BaseModel, StrictFloat, StrictInt


class Parameters(BaseModel):
    """
    Parameters to pass to the model when generating  # noqa: E501
    """

    max_tokens: Optional[StrictInt] = None
    temperature: Optional[Union[StrictFloat, StrictInt]] = None
    top_k: Optional[Union[StrictFloat, StrictInt]] = None
    top_p: Optional[Union[StrictFloat, StrictInt]] = None
    stop: Optional[Stop] = None
    __properties = ["max_tokens", "temperature", "top_k", "top_p", "stop"]

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Parameters:
        """Create an instance of Parameters from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of stop
        if self.stop:
            _dict["stop"] = self.stop.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> Parameters:
        """Create an instance of Parameters from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return Parameters.parse_obj(obj)

        _obj = Parameters.parse_obj(
            {
                "max_tokens": obj.get("max_tokens"),
                "temperature": obj.get("temperature"),
                "top_k": obj.get("top_k"),
                "top_p": obj.get("top_p"),
                "stop": Stop.from_dict(obj.get("stop"))
                if obj.get("stop") is not None
                else None,
            }
        )
        return _obj



from truefoundry.autodeploy.exception import InvalidRequirementsException


def get_git_binary():
    try:
        import git

        return git
    except Exception as ex:
        raise InvalidRequirementsException(
            message="We cannot find the 'git' command. We use Git to track changes made while automatically building your project. Please install Git to use this feature or manually create a 'truefoundry.yaml' file."
        ) from ex



import copy
import datetime
import json
import logging
import os.path
import tempfile
import uuid
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from truefoundry import client
from truefoundry.common.warnings import TrueFoundryDeprecationWarning
from truefoundry.ml._autogen.client import (  # type: ignore[attr-defined]
    ArtifactType,
    CreateArtifactVersionRequestDto,
    DeleteArtifactVersionsRequestDto,
    ExternalBlobStorageSource,
    FinalizeArtifactVersionRequestDto,
    Framework,
    MlfoundryArtifactsApi,
    ModelDto,
    ModelManifest,
    ModelVersionDto,
    ModelVersionEnvironment,
    NotifyArtifactVersionFailureDto,
    TrueFoundryManagedSource,
    UpdateModelVersionRequestDto,
)
from truefoundry.ml._autogen.models import infer_signature as _infer_signature
from truefoundry.ml.artifact.truefoundry_artifact_repo import (
    ArtifactIdentifier,
    MlFoundryArtifactsRepository,
)
from truefoundry.ml.enums import ModelFramework
from truefoundry.ml.exceptions import MlFoundryException
from truefoundry.ml.log_types.artifacts.artifact import BlobStorageDirectory
from truefoundry.ml.log_types.artifacts.constants import (
    ARTIFACT_METADATA_TRUEFOUNDRY_KEY,
    INTERNAL_METADATA_PATH,
)
from truefoundry.ml.log_types.artifacts.utils import (
    _copy_additional_files,
    _get_src_dest_pairs,
    _validate_artifact_metadata,
    _validate_description,
    calculate_total_size,
    get_autogen_type,
    set_tfy_internal_metadata,
    set_user_artifact_metadata,
)
from truefoundry.ml.model_framework import (
    ModelFrameworkType,
    _ModelFramework,
    auto_update_environment_details,
    auto_update_model_framework_details,
)
from truefoundry.ml.session import _get_api_client
from truefoundry.pydantic_v1 import BaseModel, Extra, parse_obj_as

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd

    from truefoundry.ml.mlfoundry_run import MlFoundryRun


logger = logging.getLogger(__name__)


# TODO: Support async download and upload
class ModelVersionInternalMetadata(BaseModel):
    class Config:
        extra = Extra.allow

    files_dir: str  # relative to root
    model_dir: str  # relative to `files_dir`
    model_is_null: bool = False
    framework: ModelFramework = ModelFramework.UNKNOWN
    transformers_pipeline_task: Optional[str] = None
    model_filename: Optional[str] = None
    mlfoundry_version: Optional[str] = None
    truefoundry_version: Optional[str] = None

    def dict(self, *args, **kwargs):
        dct = super().dict(*args, **kwargs)
        dct["framework"] = dct["framework"].value
        return dct


class ModelVersionDownloadInfo(BaseModel):
    download_dir: str
    model_dir: str
    model_framework: ModelFramework = ModelFramework.UNKNOWN
    model_filename: Optional[str] = None


class ModelVersion:
    def __init__(
        self,
        model_version: ModelVersionDto,
        model: ModelDto,
    ) -> None:
        self._api_client = _get_api_client()
        self._mlfoundry_artifacts_api = MlfoundryArtifactsApi(
            api_client=self._api_client
        )
        self._model_version = model_version
        self._model = model
        self._deleted = False
        self._description: str = ""
        self._metadata: Dict[str, Any] = {}
        self._environment: Optional[ModelVersionEnvironment] = None
        self._framework: Optional[ModelFrameworkType] = None
        self._version_alias: Optional[str] = None
        self._set_mutable_attrs()

    @classmethod
    def from_fqn(cls, fqn: str) -> "ModelVersion":
        """
        Get the version of a model to download contents or load them in memory

        Args:
            fqn (str): Fully qualified name of the model version.

        Returns:
            ModelVersion: An ModelVersion instance of the Model

        Examples:

            ```python
            from truefoundry.ml import get_client, ModelVersion

            client = get_client()
            model_version = ModelVersion.from_fqn(fqn="<your-model-fqn>")
            ```
        """
        api_client = _get_api_client()
        mlfoundry_artifacts_api = MlfoundryArtifactsApi(api_client=api_client)
        _model_version = mlfoundry_artifacts_api.get_model_version_by_fqn_get(fqn=fqn)
        model_version = _model_version.model_version
        _model = mlfoundry_artifacts_api.get_model_get(id=model_version.model_id)
        model = _model.model
        instance = cls(model_version=model_version, model=model)
        return instance

    def _ensure_not_deleted(self):
        if self._deleted:
            raise MlFoundryException(
                "Model Version was deleted, cannot perform updates on a deleted version"
            )

    def _set_mutable_attrs(self):
        if self._model_version.manifest:
            manifest = self._model_version.manifest
            self._description = manifest.description or ""
            self._metadata = copy.deepcopy(manifest.metadata)
            self._environment = copy.deepcopy(manifest.environment)
            self._framework = (
                parse_obj_as(
                    ModelFrameworkType, manifest.framework.actual_instance.to_dict()
                )
                if manifest.framework
                else None
            )
            self._version_alias = self._model_version.manifest.version_alias or None
        else:
            self._description = self._model_version.description or ""
            self._metadata = copy.deepcopy(self._model_version.artifact_metadata or {})
            self._environment = None
            self._framework = _ModelFramework.to_model_framework_type(
                self._model_version.model_framework
            )
            self._version_alias = None

    def _refetch_model_version(self, reset_mutable_attrs: bool = True):
        _model_version = self._mlfoundry_artifacts_api.get_model_version_get(
            id=self._model_version.id
        )
        self._model_version = _model_version.model_version
        if reset_mutable_attrs:
            self._set_mutable_attrs()

    def __repr__(self):
        return f"{self.__class__.__name__}(fqn={self.fqn!r})"

    def _get_artifacts_repo(self):
        return MlFoundryArtifactsRepository(
            artifact_identifier=ArtifactIdentifier(
                artifact_version_id=uuid.UUID(self._model_version.id)
            ),
            api_client=self._api_client,
        )

    @property
    def name(self) -> str:
        """Get the name of the model"""
        return self._model.name

    @property
    def model_fqn(self) -> str:
        """Get fqn of the model"""
        return self._model.fqn

    @property
    def version(self) -> int:
        """Get version information of the model"""
        return self._model_version.version

    @property
    def fqn(self) -> str:
        """Get fqn of the current model version"""
        return self._model_version.fqn

    @property
    def step(self) -> Optional[int]:
        """Get the step in which model was created"""
        if self._model_version.manifest:
            return self._model_version.manifest.step
        return self._model_version.step

    @property
    def description(self) -> str:
        """Get description of the model"""
        return self._description

    @description.setter
    def description(self, value: str):
        """set the description of the model"""
        _validate_description(value)
        self._description = value

    @property
    def metadata(self) -> Dict[str, Any]:
        """Get metadata for the current model"""
        return self._metadata

    @metadata.setter
    def metadata(self, value: Dict[str, Any]):
        """set the metadata for current model"""
        _validate_artifact_metadata(value)
        self._metadata = copy.deepcopy(value)

    @property
    def version_alias(self) -> Optional[str]:
        """
        Get version alias for the current model version
        """
        if not self._model_version.manifest:
            warnings.warn(
                message="This model version was created using an older serialization format. version alias does not exist",
                category=TrueFoundryDeprecationWarning,
                stacklevel=2,
            )
        return self._version_alias

    @version_alias.setter
    def version_alias(self, value: Optional[str]):
        """
        Set the version alias for current artifact version
        """
        if not self._model_version.manifest:
            warnings.warn(
                message="This model version was created using an older serialization format. version alias will not be updated",
                category=TrueFoundryDeprecationWarning,
                stacklevel=2,
            )
            return
        self._version_alias = value

    @property
    def environment(self) -> Optional[ModelVersionEnvironment]:
        """Get the environment details for the model"""
        if not self._model_version.manifest:
            warnings.warn(
                message="This model version was created using an older serialization format. environment does not exist, returning None",
                category=TrueFoundryDeprecationWarning,
                stacklevel=2,
            )
        return self._environment

    @environment.setter
    def environment(self, value: Optional[ModelVersionEnvironment]):
        """set the environment details for the model"""
        if not self._model_version.manifest:
            warnings.warn(
                message="This model version was created using an older serialization format. Environment will not be updated",
                category=TrueFoundryDeprecationWarning,
                stacklevel=2,
            )
            return
        self._environment = copy.deepcopy(value)

    @property
    def framework(self) -> Optional["ModelFrameworkType"]:
        """Get the framework of the model"""
        return self._framework

    @framework.setter
    def framework(
        self, value: Optional[Union[str, ModelFramework, "ModelFrameworkType"]]
    ):
        """Set the framework of the model"""
        if not self._model_version.manifest:
            warnings.warn(
                message="This model version was created using an older serialization format. Framework will not be updated",
                category=TrueFoundryDeprecationWarning,
                stacklevel=2,
            )
            return
        self._framework = _ModelFramework.to_model_framework_type(value)

    @property
    def created_at(self) -> Optional[datetime.datetime]:
        """Get the time at which model version was created"""
        return self._model_version.created_at

    @property
    def updated_at(self) -> Optional[datetime.datetime]:
        """Get the information about when the model version was updated"""
        return self._model_version.updated_at

    def raw_download(
        self,
        path: Optional[Union[str, Path]],
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> str:
        """
        Download a model file or directory to a local directory if applicable, and return a
        local path for it.

        Args:
            path (str): Absolute path of the local filesystem destination directory to which to
                        download the specified models. This directory must already exist.
                        If unspecified, the models will either be downloaded to a new
                        uniquely-named directory on the local filesystem.
            overwrite (bool): If True it will overwrite the file if it is already present in the download directory else
                              it will throw an error
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            path:  Absolute path of the local filesystem location containing the desired models.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            model_version = client.get_model_version_by_fqn(fqn="<your-model-fqn>")
            model_version.raw_download(path="<your-desired-download-path>")
            ```
        """
        logger.info("Downloading model version contents, this might take a while ...")
        artifacts_repo = self._get_artifacts_repo()
        return artifacts_repo.download_artifacts(
            artifact_path="", dst_path=path, overwrite=overwrite, progress=progress
        )

    def _download(
        self,
        path: Optional[Union[str, Path]],
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> ModelVersionDownloadInfo:
        self._ensure_not_deleted()
        download_dir = self.raw_download(
            path=path, overwrite=overwrite, progress=progress
        )

        if self._model_version.manifest:
            _framework = self._model_version.manifest.framework
            model_framework = (
                ModelFramework(_framework.actual_instance.type)
                if _framework
                else ModelFramework.UNKNOWN
            )
            download_info = ModelVersionDownloadInfo(
                download_dir=download_dir,
                model_dir=download_dir,
                model_framework=model_framework,
                model_filename=None,
            )
            return download_info

        internal_metadata_path = os.path.join(download_dir, INTERNAL_METADATA_PATH)
        if not os.path.exists(internal_metadata_path):
            raise MlFoundryException(
                "Model version seems to be corrupted or in invalid format due to missing model metadata. "
                "You can still use .raw_download(path='/your/path/here') to download and inspect files."
            )
        with open(internal_metadata_path) as f:
            internal_metadata = ModelVersionInternalMetadata.parse_obj(json.load(f))
        download_info = ModelVersionDownloadInfo(
            download_dir=os.path.join(download_dir, internal_metadata.files_dir),
            model_dir=os.path.join(
                download_dir,
                internal_metadata.files_dir,
                internal_metadata.model_dir,
            ),
            model_framework=internal_metadata.framework,
            model_filename=internal_metadata.model_filename,
        )
        return download_info

    def download(
        self,
        path: Optional[Union[str, Path]],
        overwrite: bool = False,
        progress: Optional[bool] = None,
    ) -> ModelVersionDownloadInfo:
        """
        Download a model file or directory to a local directory if applicable, and return download info
        containing `model_dir` - local path where model was downloaded

        Args:
            path (str): Absolute path of the local filesystem destination directory to which to
                        download the specified models. This directory must already exist.
                        If unspecified, the models will either be downloaded to a new
                        uniquely-named directory on the local filesystem.
            overwrite (bool): If True it will overwrite the file if it is already present in the download directory else
                              it will throw an error
            progress (bool): value to show progress bar, defaults to None.

        Returns:
            ModelVersionDownloadInfo:  Download Info instance containing
                `model_dir` (path to downloaded model folder) and other metadata

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            model_version = client.get_model_version_by_fqn(fqn="<your-model-fqn>")
            download_info = model_version.download(path="<your-desired-download-path>")
            print(download_info.model_dir)
            ```
        """
        download_info = self._download(
            path=path, overwrite=overwrite, progress=progress
        )
        logger.info("Downloaded model contents to %s", download_info.model_dir)
        return download_info

    def delete(self) -> bool:
        """
        Deletes the current instance of the ModelVersion hence deleting the current version.

        Returns:
            True if model was deleted successfully

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            model_version = client.get_model_version_by_fqn(fqn="<your-model-fqn>")
            model_version.delete()
            ```
        """
        self._ensure_not_deleted()
        self._mlfoundry_artifacts_api.delete_artifact_version_post(
            delete_artifact_versions_request_dto=DeleteArtifactVersionsRequestDto(
                id=self._model_version.id
            )
        )
        self._deleted = True
        return True

    def update(self):
        """
        Updates the current instance of the ModelVersion hence updating the current version.

        Examples:

            ```python
            from truefoundry.ml import get_client

            client = get_client()
            model_version = client.get_model_version_by_fqn(fqn="<your-model-fqn>")
            model_version.description = 'This is the new description'
            model_version.update()
            ```
        """
        self._ensure_not_deleted()
        if self._model_version.manifest:
            self._model_version.manifest.description = self.description
            self._model_version.manifest.metadata = self.metadata
            self._model_version.manifest.environment = self.environment
            self._model_version.manifest.framework = (
                Framework.from_dict(self.framework.dict()) if self.framework else None
            )
            self._model_version.manifest.version_alias = self.version_alias
        try:
            _model_version = self._mlfoundry_artifacts_api.update_model_version_post(
                update_model_version_request_dto=UpdateModelVersionRequestDto(
                    id=self._model_version.id,
                    description=self.description,
                    artifact_metadata=self.metadata,
                    manifest=self._model_version.manifest,
                )
            )
        except Exception:
            # rollback edits to internal object
            self._refetch_model_version(reset_mutable_attrs=False)
            raise
        else:
            self._model_version = _model_version.model_version
            self._set_mutable_attrs()


def _log_model_version(  # noqa: C901
    run: Optional["MlFoundryRun"],
    name: str,
    model_file_or_folder: Union[str, BlobStorageDirectory],
    ml_repo: Optional[str] = None,
    description: Optional[str] = None,
    metadata: Optional[Dict[str, Any]] = None,
    step: Optional[int] = 0,
    progress: Optional[bool] = None,
    framework: Optional[Union[str, ModelFramework, "ModelFrameworkType"]] = None,
    environment: Optional[ModelVersionEnvironment] = None,
) -> ModelVersion:
    if not isinstance(model_file_or_folder, (str, BlobStorageDirectory)):
        raise MlFoundryException(
            "model_file_or_folder should be of type str or BlobStorageDirectory"
        )

    if not run and not ml_repo:
        raise MlFoundryException("Exactly one of run, ml_repo should be passed")

    if run:
        mlfoundry_artifacts_api = run._mlfoundry_artifacts_api
        repos = client.ml_repos.get(id=run._experiment_id)
        ml_repo = repos.data.manifest.name
        ml_repo_id = run._experiment_id
    else:
        assert ml_repo is not None
        api_client = _get_api_client()
        mlfoundry_artifacts_api = MlfoundryArtifactsApi(api_client=api_client)
        result = list(client.ml_repos.list(name=ml_repo, limit=1))[0]
        ml_repo_id = result.id

    step = step or 0
    total_size = None
    metadata = metadata or {}
    metadata = set_tfy_internal_metadata(metadata)
    metadata = set_user_artifact_metadata(metadata)

    _validate_description(description)
    _validate_artifact_metadata(metadata)

    if isinstance(model_file_or_folder, str):
        logger.info("Logging model and additional files, this might take a while ...")
        temp_dir = tempfile.TemporaryDirectory(prefix="truefoundry-")
        try:
            logger.info("Adding model file/folder to model version content")
            temp_dest_to_src_map = _copy_additional_files(
                root_dir=temp_dir.name,
                files_dir="",
                model_dir=None,
                additional_files=[(model_file_or_folder, "")],
                ignore_model_dir_dest_conflict=True,
            )

        except Exception as e:
            temp_dir.cleanup()
            raise MlFoundryException("Failed to log model") from e

    # create entry
    _create_artifact_version_response = (
        mlfoundry_artifacts_api.create_artifact_version_post(
            create_artifact_version_request_dto=CreateArtifactVersionRequestDto(
                experiment_id=int(ml_repo_id),
                artifact_type=ArtifactType.MODEL,
                name=name,
            )
        )
    )
    version_id = _create_artifact_version_response.id
    artifact_storage_root = _create_artifact_version_response.artifact_storage_root
    if isinstance(model_file_or_folder, str):
        # Source is of type TrueFoundryManagedSource
        source = TrueFoundryManagedSource(type="truefoundry", uri=artifact_storage_root)
        artifacts_repo = MlFoundryArtifactsRepository(
            artifact_identifier=ArtifactIdentifier(
                artifact_version_id=uuid.UUID(version_id)
            ),
            api_client=mlfoundry_artifacts_api.api_client,
        )
        total_size = calculate_total_size(list(temp_dest_to_src_map.values()))
        try:
            logger.info(
                "Packaging and uploading files to remote with Total Size: %.6f MB",
                total_size / 1000000.0,
            )
            src_dest_pairs = _get_src_dest_pairs(
                root_dir=temp_dir.name, dest_to_src_map=temp_dest_to_src_map
            )
            artifacts_repo.log_artifacts(
                src_dest_pairs=src_dest_pairs, progress=progress
            )
        except Exception as e:
            mlfoundry_artifacts_api.notify_failure_post(
                notify_artifact_version_failure_dto=NotifyArtifactVersionFailureDto(
                    id=version_id
                )
            )
            raise MlFoundryException("Failed to log model") from e
        finally:
            temp_dir.cleanup()
    elif isinstance(model_file_or_folder, BlobStorageDirectory):
        source = ExternalBlobStorageSource(
            type="external", uri=model_file_or_folder.uri
        )
    else:
        raise MlFoundryException("Invalid model_file_or_folder provided")

    if total_size is not None and total_size > 0:
        metadata[ARTIFACT_METADATA_TRUEFOUNDRY_KEY]["artifact_size"] = total_size

    _source_cls = get_autogen_type(ModelManifest, "source")
    # Auto fetch the framework & environment details if not provided
    framework = _ModelFramework.to_model_framework_type(framework)
    if framework and isinstance(model_file_or_folder, str):
        auto_update_model_framework_details(
            framework=framework, model_file_or_folder=model_file_or_folder
        )
    environment = environment or ModelVersionEnvironment()
    auto_update_environment_details(environment=environment, framework=framework)

    model_manifest = ModelManifest(
        name=name,
        ml_repo=ml_repo,
        description=description,
        metadata=metadata,
        source=_source_cls.from_dict(source.dict()),
        framework=Framework.from_dict(framework.dict()) if framework else None,
        environment=environment,
        step=step,
        run_id=run.run_id if run else None,
    )
    _manifest_cls = get_autogen_type(FinalizeArtifactVersionRequestDto, "manifest")
    artifact_version_response = mlfoundry_artifacts_api.finalize_artifact_version_post(
        finalize_artifact_version_request_dto=FinalizeArtifactVersionRequestDto(
            id=version_id,
            run_uuid=run.run_id if run else None,
            artifact_size=total_size,
            artifact_metadata=metadata,
            internal_metadata=None,
            step=model_manifest.step,
            manifest=_manifest_cls.from_dict(model_manifest.to_dict()),
        )
    )
    return ModelVersion.from_fqn(fqn=artifact_version_response.artifact_version.fqn)


def infer_signature(
    model_input: Any = None,
    model_output: Optional[
        Union["pd.DataFrame", "np.ndarray", Dict[str, "np.ndarray"]]
    ] = None,
    params: Optional[Dict[str, Any]] = None,
):
    return _infer_signature(
        model_input=model_input, model_output=model_output, params=params
    )



import os
import shutil
from typing import Any, Callable, Dict, Optional

from flytekit import FlyteContextManager, PythonFunctionTask, lazy_module
from flytekit.configuration import SerializationSettings
from flytekit.core.context_manager import ExecutionParameters
from flytekit.extend import ExecutionState, TaskPlugins
from flytekit.extend.backend.base_agent import AsyncAgentExecutorMixin

from truefoundry.deploy.v2.lib.patched_models import PySparkTaskConfig

pyspark_sql = lazy_module("pyspark.sql")
SparkSession = pyspark_sql.SparkSession


class TfySparkFunctionTask(
    AsyncAgentExecutorMixin, PythonFunctionTask[PySparkTaskConfig]
):
    """
    Actual Plugin that transforms the local python code for execution within a spark context
    """

    _SPARK_TASK_TYPE = "spark"

    def __init__(
        self,
        task_config: PySparkTaskConfig,
        task_function: Callable,
        **kwargs,
    ):
        self.sess: Optional[SparkSession] = None  # type: ignore

        task_type = self._SPARK_TASK_TYPE

        super(TfySparkFunctionTask, self).__init__(
            task_config=task_config,
            task_type=task_type,
            task_function=task_function,
            **kwargs,
        )

    def get_custom(self, settings: SerializationSettings) -> Dict[str, Any]:
        return {"truefoundry": self._task_config.dict()}

    def pre_execute(self, user_params: ExecutionParameters) -> ExecutionParameters:
        import pyspark as _pyspark

        ctx = FlyteContextManager.current_context()
        sess_builder = _pyspark.sql.SparkSession.builder.appName(
            f"FlyteSpark: {user_params.execution_id}"
        )
        if not (
            ctx.execution_state
            and ctx.execution_state.mode == ExecutionState.Mode.TASK_EXECUTION
        ):
            # If either of above cases is not true, then we are in local execution of this task
            # Add system spark-conf for local/notebook based execution.
            spark_conf = _pyspark.SparkConf()
            spark_conf.set("spark.driver.bindAddress", "127.0.0.1")
            for k, v in self.task_config.spark_conf.items():
                spark_conf.set(k, v)
            # In local execution, propagate PYTHONPATH to executors too. This makes the spark
            # execution hermetic to the execution environment. For example, it allows running
            # Spark applications using Bazel, without major changes.
            if "PYTHONPATH" in os.environ:
                spark_conf.setExecutorEnv("PYTHONPATH", os.environ["PYTHONPATH"])
            sess_builder = sess_builder.config(conf=spark_conf)

        self.sess = sess_builder.getOrCreate()

        if (
            ctx.serialization_settings
            and ctx.serialization_settings.fast_serialization_settings
            and ctx.serialization_settings.fast_serialization_settings.enabled
            and ctx.execution_state
            and ctx.execution_state.mode == ExecutionState.Mode.TASK_EXECUTION
        ):
            file_name = "flyte_wf"
            file_format = "zip"
            shutil.make_archive(file_name, file_format, os.getcwd())
            self.sess.sparkContext.addPyFile(f"{file_name}.{file_format}")

        return user_params.builder().add_attr("SPARK_SESSION", self.sess).build()

    def execute(self, **kwargs) -> Any:
        return PythonFunctionTask.execute(self, **kwargs)


# Inject the Spark plugin into flytekits dynamic plugin loading system
TaskPlugins.register_pythontask_plugin(PySparkTaskConfig, TfySparkFunctionTask)



from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import git as gitpython
import yaml

from truefoundry.cli.console import console
from truefoundry.deploy.lib.clients.servicefoundry_client import (
    ServiceFoundryServiceClient,
)
from truefoundry.deploy.lib.diff_utils import print_manifest_diff
from truefoundry.deploy.lib.messages import PROMPT_APPLYING_MANIFEST
from truefoundry.deploy.lib.model.entity import (
    ApplyResult,
    DependencyTreeResponse,
    File,
    GitChangedFiles,
    GitChangedYamlFiles,
    ManifestLike,
    ResolvedResourceInfo,
    ResolvedResources,
)
from truefoundry.pydantic_v1 import ValidationError

MANIFEST_TYPE_ORDER = [
    "team",
    "provider-account",
    "environment",
    "cluster",
    "ml-repo",
    "workspace",
    "service",
    "job",
    "helm",
    "notebook",
    "rstudio",
    "volume",
    "workflow",
    "async-service",
    "application-set",
    "ssh-server",
    "spark-job",
    "secret-group",
    "tracing-project",
    "mcp-server",
    "agent",
    "policy",
    "role",
    "virtual-account",
    "gateway-guardrails-config",
    "gateway-rate-limiting-config",
    "gateway-global-settings",
    "gateway-fallback-config",
    "gateway-load-balancing-config",
    "gateway-budget-config",
    "gateway-data-access-config",
    "gateway-otel-config",
    "gateway-data-routing-config",
    "gateway-installation",
    "settings",
    "alert-config",
]

_UNKNOWN_TYPE_PRIORITY = len(MANIFEST_TYPE_ORDER)


def _get_base_type(type_str: str) -> str:
    """Extract base type, e.g. 'provider-account/gcp' -> 'provider-account'."""
    return type_str.split("/")[0]


def _get_type_priority(type_str: str) -> int:
    base_type = _get_base_type(type_str)
    try:
        return MANIFEST_TYPE_ORDER.index(base_type)
    except ValueError:
        return _UNKNOWN_TYPE_PRIORITY


def get_type_and_name_from_manifest(
    manifest: Dict[str, Any],
) -> Optional[ManifestLike]:
    if manifest and isinstance(manifest.get("type"), str):
        type_str = manifest.get("type")
        # Fallback to type if name is not present in the manifest
        name_str = manifest.get("name") or type_str
        return ManifestLike(type=type_str, name=name_str)
    return None


def read_manifest_type_and_name(filepath: str) -> Optional[ManifestLike]:
    """Read the type and name fields from a manifest file."""
    try:
        manifest = get_manifest_from_file(filepath)
        return get_type_and_name_from_manifest(manifest)
    except Exception:
        pass
    return None


def collect_yaml_files(directory: str) -> List[str]:
    """Collect all .yaml and .yml files from a directory recursively."""
    dir_path = Path(directory).resolve()
    yaml_files: List[str] = []
    for ext in ("*.yaml", "*.yml"):
        yaml_files.extend(str(p) for p in dir_path.rglob(ext))
    return yaml_files


def sort_files_by_manifest_type(files: List[str]) -> List[str]:
    """Sort files by their manifest type according to dependency order."""

    def sort_key(filepath: str) -> Tuple[int, str]:
        manifest_type_and_name = read_manifest_type_and_name(filepath)
        if manifest_type_and_name is None:
            return (_UNKNOWN_TYPE_PRIORITY + 1, filepath)
        return (_get_type_priority(manifest_type_and_name.type), filepath)

    return sorted(files, key=sort_key)


def _open_git_repo(directory: str) -> gitpython.Repo:
    """Open a git repository, raising clear errors for common failure cases."""
    try:
        repo = gitpython.Repo(directory, search_parent_directories=True)
    except gitpython.InvalidGitRepositoryError:
        raise Exception(f"{directory} is not inside a git repository") from None
    except gitpython.NoSuchPathError:
        raise Exception(f"Directory {directory} does not exist") from None
    except gitpython.GitCommandNotFound:
        raise Exception("git is not installed or not found in PATH") from None

    if repo.working_tree_dir is None:
        raise Exception("Bare git repositories are not supported")
    return repo


def _get_changed_files(repo: gitpython.Repo, ref: str) -> GitChangedFiles:
    """Get all modified file paths (tracked + untracked), deleted file paths compared to a ref."""
    modified_files: List[File] = []
    deleted_files: List[File] = []

    try:
        for diff_item in repo.commit(ref).diff(None):
            file_path = diff_item.a_path or diff_item.b_path
            if diff_item.change_type == "D":
                blob_text = ""
                if file_path and diff_item.a_blob:
                    try:
                        blob_text = diff_item.a_blob.data_stream.read().decode(
                            "utf-8", errors="replace"
                        )
                    except Exception:
                        pass
                deleted_files.append(File(path=file_path, content=blob_text))
            else:
                modified_files.append(File(path=file_path))
    except Exception as e:
        raise Exception(f"Failed to compute diff against '{ref}': {e}") from None

    for untracked_file in repo.untracked_files:
        modified_files.append(File(path=untracked_file))
    return GitChangedFiles(modified=modified_files, deleted=deleted_files)


def _get_git_changed_yaml_file_path_if_valid(
    f: str, repo_root: Path, dir_path: Path
) -> Union[Path, None]:
    if not f.endswith((".yaml", ".yml")):
        return None
    abs_path = (repo_root / f).resolve()
    try:
        abs_path.relative_to(dir_path)
    except ValueError:
        return None
    return abs_path


def get_git_changed_yaml_files(
    directory: str, ref: Optional[str] = None
) -> GitChangedYamlFiles:
    """Get YAML files changed in git compared to ref (default: HEAD).

    Includes both tracked changes and untracked (new) files, filtered to
    only those within the given directory.
    """
    dir_path = Path(directory).resolve()
    ref = ref or "HEAD"

    repo = _open_git_repo(str(dir_path))
    repo_root = Path(repo.working_tree_dir)
    changed_files_info = _get_changed_files(repo, ref)

    git_modified_yaml_files: List[File] = []
    for modified_file in changed_files_info.modified:
        abs_path = _get_git_changed_yaml_file_path_if_valid(
            modified_file.path, repo_root, dir_path
        )
        if abs_path and abs_path.is_file():
            git_modified_yaml_files.append(File(path=str(abs_path), content=None))

    git_deleted_yaml_files: List[File] = []
    for deleted_file in changed_files_info.deleted:
        abs_path = _get_git_changed_yaml_file_path_if_valid(
            deleted_file.path, repo_root, dir_path
        )
        if abs_path:
            git_deleted_yaml_files.append(
                File(path=str(abs_path), content=deleted_file.content)
            )

    return GitChangedYamlFiles(
        modified=git_modified_yaml_files,
        deleted=git_deleted_yaml_files,
    )


def _apply_manifest(
    manifest: Dict[str, Any],
    client: Optional[ServiceFoundryServiceClient] = None,
    filename: Optional[str] = None,
    index: Optional[int] = None,
    dry_run: bool = False,
    show_diff: bool = False,
) -> ApplyResult:
    client = client or ServiceFoundryServiceClient()

    file_metadata = ""
    if index is not None:
        file_metadata += f" at index {index}"
    if filename:
        file_metadata += f" from file {filename}"

    try:
        parsed_manifest = ManifestLike.parse_obj(manifest)
    except ValidationError as ex:
        return ApplyResult(
            success=False,
            message=f"Failed to apply manifest{file_metadata}. Error: {ex}",
        )

    prefix = "[Dry Run] " if dry_run else ""
    suffix = " (No changes were applied)" if dry_run else ""

    manifest_name = ""
    if parsed_manifest.name:
        manifest_name = f" {parsed_manifest.name}"

    try:
        api_response = client.apply(
            parsed_manifest.dict(exclude_unset=True, exclude_none=True), dry_run
        )

        # Show diff for dry runs only when show_diff is enabled
        if dry_run and show_diff and api_response.existing_manifest:
            if parsed_manifest.name:
                diff_manifest_name = f"{parsed_manifest.name} ({parsed_manifest.type})"
            else:
                diff_manifest_name = f"manifest ({parsed_manifest.type})"
            print_manifest_diff(
                existing_manifest=api_response.existing_manifest,
                new_manifest=parsed_manifest.dict(
                    exclude_unset=True, exclude_none=True
                ),
                manifest_name=diff_manifest_name,
                console=console,
            )

        return ApplyResult(
            success=True,
            message=(
                f"{prefix}Successfully configured manifest{manifest_name} of type {parsed_manifest.type}.{suffix}"
            ),
        )
    except Exception as ex:
        return ApplyResult(
            success=False,
            message=(
                f"{prefix}Failed to apply manifest{manifest_name} of type {parsed_manifest.type}. Error: {ex}.{suffix}"
            ),
        )


def apply_manifest(
    manifest: Dict[str, Any],
    client: Optional[ServiceFoundryServiceClient] = None,
    dry_run: bool = False,
    show_diff: bool = False,
) -> ApplyResult:
    return _apply_manifest(
        manifest=manifest, client=client, dry_run=dry_run, show_diff=show_diff
    )


def get_manifest_from_file(filepath: str) -> Optional[Dict[str, Any]]:
    try:
        with open(filepath, "r") as f:
            manifest = yaml.safe_load(f)
            if isinstance(manifest, dict):
                return manifest
    except Exception:
        pass
    return None


def apply_manifest_file(
    filepath: str,
    client: Optional[ServiceFoundryServiceClient] = None,
    dry_run: bool = False,
    show_diff: bool = False,
) -> Iterator[ApplyResult]:
    client = client or ServiceFoundryServiceClient()
    filename = Path(filepath).name
    try:
        with open(filepath, "r") as f:
            manifests_it = list(yaml.safe_load_all(f))
    except Exception as ex:
        yield ApplyResult(
            success=False,
            message=f"Failed to read file {filepath} as a valid YAML file. Error: {ex}",
        )
    else:
        prefix = "[Dry Run] " if dry_run else ""
        for index, manifest in enumerate(manifests_it):
            if not isinstance(manifest, dict):
                yield ApplyResult(
                    success=False,
                    message=f"{prefix}Failed to apply manifest at index {index} from file {filename}. Error: A manifest must be a dict, got type {type(manifest)}",
                )
                continue

            yield _apply_manifest(
                manifest=manifest,
                client=client,
                filename=filename,
                index=index,
                dry_run=dry_run,
                show_diff=show_diff,
            )


def apply_manifest_files(
    files: List[str],
    sf_server_client: Optional[ServiceFoundryServiceClient] = None,
    dry_run: bool = False,
    show_diff: bool = False,
) -> List[ApplyResult]:
    apply_results: List[ApplyResult] = []
    for file in files:
        with console.status(PROMPT_APPLYING_MANIFEST.format(file), spinner="dots"):
            for apply_result in apply_manifest_file(
                file, sf_server_client, dry_run, show_diff
            ):
                if apply_result.success:
                    console.print(f"[green]\u2714 {apply_result.message}[/]")
                else:
                    console.print(f"[red]\u2718 {apply_result.message}[/]")

                apply_results.append(apply_result)
    return apply_results


def _get_type_name_key(type: str, name: str) -> str:
    return f"{type}:{name}"


def resolve_resources_for_dry_run(
    given_files: List[str], client: ServiceFoundryServiceClient
) -> ResolvedResources:
    # Map each <type:name> pair to the filepath in given_files list
    type_name_to_filepath: Dict[str, str] = {}
    # Manifests to check for dependencies
    manifests_to_check: List[Dict[str, Any]] = []
    # Type and name pairs list
    type_and_name_list: List[ManifestLike] = []

    for filepath in given_files:
        manifest = get_manifest_from_file(filepath)
        if manifest:
            type_and_name = get_type_and_name_from_manifest(manifest)
            if type_and_name:
                type_and_name_list.append(type_and_name)
                type_name_to_filepath[
                    _get_type_name_key(type_and_name.type, type_and_name.name)
                ] = filepath
                manifests_to_check.append(manifest)

    # Extract the processable and skipped resources from the dependency tree
    processable_resources: List[ResolvedResourceInfo] = []
    skipped_resources: List[ResolvedResourceInfo] = []

    dependency_tree_response: Optional[DependencyTreeResponse] = None
    try:
        dependency_tree_response = client.resolve_dependency_tree(manifests_to_check)
    except Exception:
        dependency_tree_response = None

    if dependency_tree_response is not None:
        for resource in dependency_tree_response.data:
            type_name_key = _get_type_name_key(resource.type, resource.name)
            filepath = type_name_to_filepath[type_name_key]
            resolved_resource_info = ResolvedResourceInfo(
                type=resource.type,
                name=resource.name,
                filepath=filepath,
            )
            if resource.dependencies and len(resource.dependencies) > 0:
                skipped_resources.append(resolved_resource_info)
            else:
                processable_resources.append(resolved_resource_info)
    else:
        for type_and_name in type_and_name_list:
            processable_resources.append(
                ResolvedResourceInfo(
                    type=type_and_name.type,
                    name=type_and_name.name,
                    filepath=type_name_to_filepath[
                        _get_type_name_key(type_and_name.type, type_and_name.name)
                    ],
                )
            )

    return ResolvedResources(
        processable_resources=processable_resources, skipped_resources=skipped_resources
    )



from __future__ import annotations

from typing import Dict, Optional

import docker
from openai import OpenAI

from truefoundry.autodeploy.agents.base import Agent
from truefoundry.autodeploy.agents.project_identifier import ProjectIdentifier
from truefoundry.autodeploy.agents.tester import Tester
from truefoundry.autodeploy.tools import (
    Ask,
    Commit,
    DockerBuild,
    ListFiles,
    ReadFile,
    RequestEvent,
    ResponseEvent,
)
from truefoundry.pydantic_v1 import Field


class Developer(Agent):
    system_prompt = """
You are a software engineer.
You goal is to do whatever you have to do to successfully run a project.
If the project already contains a dockerfile, you can use that.
If the project does not contain a dockerfile, you have to create one.
You need to fix any issue to ensure that the image builds succesfully.
Generate syntactically correct Dockerfile.
You need to always read a file before making any changes.
Ensure that you always make changes if you are writing a file.
This is the rough flow you have to follow:
Identify Project -> Build Docker Image -> Tester
If Tester is not successful, you need to fix the issue and try Tester again.
Tester has the final authority over whether you were successful in your task.
Do not ask the user to manually review anything. You need to review and and take a decision.
You can ask questions to the user but you need to take actions to run the project.
Do not ask users how to fix an issue.
Communicate with the user only via Ask tool.
You should add Dockerfile path to .dockerignore file, this will speed up iteration.
You can create a .dockerignore file if it does not exist.
Try to download dependencies first, before you copy the whole project. This will speed up
subsequent builds due to build cache.
Use the Project identifier tool call to identify project.
Avoid reading *.lock type files as they tend to be large
If you are using multi-stage build:
1. For a go project  the final stage should not need a go runtime.
2. If you get a missing file error because the application is trying to load a file, you should
   add the files in the final stage.
3. While copying the file, be careful about the destination directory.

For Golang projects:
1. you need to ensure you know where is the main.go file. Sometimes it can be in cmd dir.

    """
    max_iter = 50

    class Request(RequestEvent):
        name: str = Field(..., description="Name of the project.")
        command: Optional[str] = Field(
            None,
            description="Preferred command to run project. This can be corrected later.",
        )

    class Response(ResponseEvent):
        command: str = Field(
            description="""
Final command to run the project within the container.
This command should be same as what is expressed uning
the entrypoint and cmd of the dockerfile.
        """
        )
        dockerfile_path: str = Field(description="Path of dockerfile.")
        port: Optional[int] = Field(description="Port in which container is running.")
        justification: str = Field(
            ...,
            description="""
            Why did you send the response back?
        """,
        )

    def __init__(
        self,
        project_root_path: str,
        docker_client: docker.DockerClient,
        openai_client: OpenAI,
        environment: Dict,
    ):
        self.tools = [
            DockerBuild(
                project_root_path=project_root_path,
                docker_client=docker_client,
            ),
            ReadFile(
                project_root_path=project_root_path,
            ),
            Commit(project_root_path=project_root_path),
            ListFiles(
                project_root_path=project_root_path,
            ),
            ProjectIdentifier(
                project_root_path=project_root_path,
                openai_client=openai_client,
            ),
            Tester(
                openai_client=openai_client,
                docker_client=docker_client,
                environment=environment,
            ),
            Ask(),
        ]
        self.openai_client = openai_client



import time
from typing import Optional

import jwt
from typing_extensions import NotRequired, TypedDict

from truefoundry.pydantic_v1 import BaseModel, Field, NonEmptyStr, validator


class UserInfo(BaseModel):
    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    user_id: NonEmptyStr
    email: Optional[str] = None
    tenant_name: NonEmptyStr = Field(alias="tenantName")


class _DecodedToken(TypedDict):
    tenantName: str
    exp: int
    username: NotRequired[str]
    email: NotRequired[str]


def _user_slug(decoded_token: _DecodedToken) -> str:
    return (
        decoded_token.get("username")
        or decoded_token.get("email")
        or "--user-slug-missing--"
    )


class Token(BaseModel):
    access_token: NonEmptyStr = Field(alias="accessToken", repr=False)
    refresh_token: Optional[NonEmptyStr] = Field(alias="refreshToken", repr=False)
    decoded_value: Optional[_DecodedToken] = Field(exclude=True, repr=False)

    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    @validator("decoded_value", always=True, pre=True)
    def _decode_jwt(cls, v, values, **kwargs) -> _DecodedToken:
        access_token = values["access_token"]
        return jwt.decode(
            access_token,
            options={
                "verify_signature": False,
                "verify_aud": False,
                "verify_exp": False,
            },
        )

    @property
    def tenant_name(self) -> str:
        assert self.decoded_value is not None
        return self.decoded_value["tenantName"]

    @property
    def exp(self) -> int:
        assert self.decoded_value is not None
        return self.decoded_value["exp"]

    def is_going_to_be_expired(self, buffer_in_seconds: int = 120) -> bool:
        assert self.decoded_value is not None
        return (self.exp - time.time()) < buffer_in_seconds

    def to_user_info(self) -> UserInfo:
        assert self.decoded_value is not None
        return UserInfo(
            user_id=_user_slug(self.decoded_value),
            email=self.decoded_value["email"]
            if "email" in self.decoded_value
            else None,
            tenant_name=self.tenant_name,
        )


class CredentialsFileContent(BaseModel):
    class Config:
        allow_mutation = False

    access_token: NonEmptyStr = Field(repr=False)
    refresh_token: Optional[NonEmptyStr] = Field(repr=False)
    host: NonEmptyStr

    def to_token(self) -> Token:
        return Token(access_token=self.access_token, refresh_token=self.refresh_token)


class TenantInfo(BaseModel):
    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    tenant_name: NonEmptyStr = Field(alias="tenantName")
    auth_server_url: str


class PythonSDKConfig(BaseModel):
    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    min_version: str = Field(alias="minVersion")
    truefoundry_cli_min_version: str = Field(alias="truefoundryCliMinVersion")
    use_sfy_server_auth_apis: Optional[bool] = Field(
        alias="useSFYServerAuthAPIs", default=False
    )
    python_build_default_image_tag: str = Field(
        alias="pythonBuildDefaultImageTag", default="3.11"
    )


class DeviceCode(BaseModel):
    class Config:
        allow_population_by_field_name = True
        allow_mutation = False

    user_code: str = Field(alias="userCode")
    device_code: str = Field(alias="deviceCode")
    verification_url: Optional[str] = Field(alias="verificationURI")
    complete_verification_url: Optional[str] = Field(alias="verificationURIComplete")
    expires_in_seconds: int = Field(alias="expiresInSeconds", default=60)
    interval_in_seconds: int = Field(alias="intervalInSeconds", default=1)
    message: Optional[str] = Field(alias="message")

    def get_user_clickable_url(self, auth_host: str) -> str:
        return f"{auth_host}/authorize/device?userCode={self.user_code}"



import enum
import os
import re
import warnings
from typing import Literal, Optional, Union

from truefoundry.common.warnings import TrueFoundryDeprecationWarning
from truefoundry.deploy._autogen import models
from truefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    constr,
    root_validator,
    validator,
)

AUTO_DISCOVERED_REQUIREMENTS_TXT_WARNING_MESSAGE_TEMPLATE = """\
Using automatically discovered {requirements_txt_path} as the requirements file.
This auto discovery behavior is deprecated and will be removed in a future release.
Please specify the relative path of requirements file explicitly as

```python
build=Build(
    ...
    build_spec=PythonBuild(
        ...
        python_dependencies=Pip(requirements_path={requirements_txt_path!r})
    )
)
```

OR

```yaml
build:
  type: build
  build_spec:
    type: tfy-python-buildpack
    python_dependencies:
      type: pip
      requirements_path: {requirements_txt_path!r}
    ...
...
```

or set it to None if you don't want to use any requirements file.
"""

SPECS_UPGRADE_WARNING_MESSAGE_TEMPLATE = """\
The `requirements_path` and `pip_packages` fields are deprecated.
It is recommended to use the `python_dependencies` field instead which supports pip, uv and poetry.
Please use the following format:

```python
build=Build(
    ...
    build_spec=PythonBuild(
        ...
        python_dependencies=Pip(requirements_path={requirements_txt_path!r}, pip_packages={pip_packages!r})
    )
)
```

OR

```yaml
build:
  type: build
  build_spec:
    type: tfy-python-buildpack
    python_dependencies:
      type: pip
      requirements_path: {requirements_txt_path!r}
      pip_packages: {pip_packages!r}
    ...
...
```
"""


def _resolve_requirements_path(
    build_context_path: str,
    requirements_path: Optional[str],
) -> Optional[str]:
    if requirements_path:
        return requirements_path

    # TODO: Deprecated behavior, phase out auto discovery in future release
    possible_requirements_txt_filename = "requirements.txt"
    possible_requirements_txt_path = os.path.join(
        build_context_path, possible_requirements_txt_filename
    )

    if os.path.isfile(possible_requirements_txt_path):
        requirements_txt_path = os.path.relpath(
            possible_requirements_txt_path, start=build_context_path
        )
        warnings.warn(
            AUTO_DISCOVERED_REQUIREMENTS_TXT_WARNING_MESSAGE_TEMPLATE.format(
                requirements_txt_path=requirements_txt_path
            ),
            category=TrueFoundryDeprecationWarning,
            stacklevel=2,
        )
        return requirements_txt_path

    return None


class CUDAVersion(str, enum.Enum):
    CUDA_11_0_CUDNN8 = "11.0-cudnn8"
    CUDA_11_1_CUDNN8 = "11.1-cudnn8"
    CUDA_11_2_CUDNN8 = "11.2-cudnn8"
    CUDA_11_3_CUDNN8 = "11.3-cudnn8"
    CUDA_11_4_CUDNN8 = "11.4-cudnn8"
    CUDA_11_5_CUDNN8 = "11.5-cudnn8"
    CUDA_11_6_CUDNN8 = "11.6-cudnn8"
    CUDA_11_7_CUDNN8 = "11.7-cudnn8"
    CUDA_11_8_CUDNN8 = "11.8-cudnn8"
    CUDA_12_0_CUDNN8 = "12.0-cudnn8"
    CUDA_12_1_CUDNN8 = "12.1-cudnn8"
    CUDA_12_2_CUDNN8 = "12.2-cudnn8"
    CUDA_12_3_CUDNN9 = "12.3-cudnn9"
    CUDA_12_4_CUDNN9 = "12.4-cudnn9"
    CUDA_12_5_CUDNN9 = "12.5-cudnn9"
    CUDA_12_6_CUDNN9 = "12.6-cudnn9"
    CUDA_12_8_CUDNN9 = "12.8-cudnn9"
    CUDA_12_9_CUDNN9 = "12.9-cudnn9"


class GPUType(str, enum.Enum):
    P4 = "P4"
    P100 = "P100"
    V100 = "V100"
    T4 = "T4"
    A10G = "A10G"
    A10_4GB = "A10_4GB"
    A10_8GB = "A10_8GB"
    A10_12GB = "A10_12GB"
    A10_24GB = "A10_24GB"
    A100_40GB = "A100_40GB"
    A100_80GB = "A100_80GB"
    L4 = "L4"
    L40S = "L40S"
    H100_80GB = "H100_80GB"
    H100_94GB = "H100_94GB"
    H100_96GB = "H100_96GB"
    H200 = "H200"
    B200 = "B200"


class TPUType(str, enum.Enum):
    V4_PODSLICE = "tpu-v4-podslice"
    V5_LITE_DEVICE = "tpu-v5-lite-device"
    V5_LITE_PODSLICE = "tpu-v5-lite-podslice"
    V5P_SLICE = "tpu-v5p-slice"


class AWSInferentiaAccelerator(str, enum.Enum):
    INF1 = "INF1"
    INF2 = "INF2"


class PatchedModelBase(BaseModel):
    class Config:
        extra = "forbid"


class DockerFileBuild(models.DockerFileBuild, PatchedModelBase):
    type: Literal["dockerfile"] = "dockerfile"

    @validator("build_args")
    def validate_build_args(cls, value):
        if not isinstance(value, dict):
            raise TypeError("build_args should be of type dict")
        for k, v in value.items():
            if not isinstance(k, str) or not isinstance(v, str):
                raise TypeError("build_args should have keys and values as string")
            if not k.strip() or not v.strip():
                raise ValueError("build_args cannot have empty keys or values")
        return value


class PythonBuild(models.PythonBuild, PatchedModelBase):
    type: Literal["tfy-python-buildpack"] = "tfy-python-buildpack"

    @root_validator
    def validate_values(cls, values):
        if values.get("cuda_version"):
            python_version = values.get("python_version")
            if python_version and not re.match(r"^3\.\d+$", python_version):
                raise ValueError(
                    f'`python_version` must be 3.x (e.g. "3.9") when `cuda_version` field is '
                    f"provided but got {python_version!r}. If you are adding a "
                    f'patch version, please remove it (e.g. "3.9.2" should be "3.9")'
                )
        _resolve_requirements_path(
            build_context_path=values.get("build_context_path") or "./",
            requirements_path=values.get("requirements_path"),
        )

        if (
            values.get("requirements_path") or values.get("pip_packages")
        ) and not values.get("python_dependencies"):
            warnings.warn(
                SPECS_UPGRADE_WARNING_MESSAGE_TEMPLATE.format(
                    requirements_txt_path=values.get("requirements_path"),
                    pip_packages=values.get("pip_packages"),
                ),
                category=TrueFoundryDeprecationWarning,
                stacklevel=2,
            )
        return values


class SparkBuild(models.SparkBuild, PatchedModelBase):
    type: Literal["tfy-spark-buildpack"] = "tfy-spark-buildpack"


class SparkImageBuild(models.SparkImageBuild, PatchedModelBase):
    type: Literal["spark-image-build"] = "spark-image-build"


class SparkImage(models.SparkImage, PatchedModelBase):
    type: Literal["spark-image"] = "spark-image"


class RemoteSource(models.RemoteSource, PatchedModelBase):
    type: Literal["remote"] = "remote"


class LocalSource(models.LocalSource, PatchedModelBase):
    type: Literal["local"] = "local"


class Build(models.Build, PatchedModelBase):
    type: Literal["build"] = "build"
    build_source: Union[models.RemoteSource, models.GitSource, models.LocalSource] = (
        Field(default_factory=LocalSource)
    )


class Manual(models.Manual, PatchedModelBase):
    type: Literal["manual"] = "manual"


class Schedule(models.Schedule, PatchedModelBase):
    type: Literal["scheduled"] = "scheduled"


class GitSource(models.GitSource, PatchedModelBase):
    type: Literal["git"] = "git"


class HttpProbe(models.HttpProbe, PatchedModelBase):
    type: Literal["http"] = "http"


class BasicAuthCreds(models.BasicAuthCreds, PatchedModelBase):
    type: Literal["basic_auth"] = "basic_auth"


class JwtAuthConfig(models.JwtAuthConfig, PatchedModelBase):
    type: Literal["jwt_auth"] = "jwt_auth"


class TrueFoundryInteractiveLogin(models.TrueFoundryInteractiveLogin, PatchedModelBase):
    type: Literal["truefoundry_oauth"] = "truefoundry_oauth"


class HealthProbe(models.HealthProbe, PatchedModelBase):
    pass


class Image(models.Image, PatchedModelBase):
    type: Literal["image"] = "image"


class Port(models.Port, PatchedModelBase):
    app_protocol: Optional[models.AppProtocol] = Field(
        "http",
        description="+label=Application Protocol\n+usage=Application Protocol for the port.\nSelect the application protocol used by your service. For most use cases, this should be `http`(HTTP/1.1).\nIf you are running a gRPC server, select the `grpc` option.\nThis is only applicable if `expose=true`.",
    )

    @root_validator(pre=True)
    def app_protocol_set_default(cls, values):
        if values.get("app_protocol", None) is None:
            values["app_protocol"] = models.AppProtocol.http.value
        return values

    @root_validator(pre=True)
    def verify_host(cls, values):
        expose = values.get("expose", True)
        host = values.get("host", None)
        if expose:
            if not host:
                raise ValueError("Host must be provided to expose port")
            if not (
                re.fullmatch(
                    r"^((([a-zA-Z0-9\-]{1,63}\.)([a-zA-Z0-9\-]{1,63}\.)*([A-Za-z]{1,63}))|(((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)))$",
                    host,
                )
            ):
                raise ValueError(
                    "Invalid value for `host`. A valid host must contain only alphanumeric letters and hypens e.g.  `ai.example.com`, `app.truefoundry.com`.\nYou can get the list of configured hosts for the cluster from the Integrations > Clusters page. Please see https://docs.truefoundry.com/docs/checking-configured-domain for more information."
                )
        else:
            if host:
                raise ValueError("Cannot pass host when expose=False")

        return values


class Resources(models.Resources, PatchedModelBase):
    pass


class Param(models.Param, PatchedModelBase):
    pass


class CPUUtilizationMetric(models.CPUUtilizationMetric, PatchedModelBase):
    type: Literal["cpu_utilization"] = "cpu_utilization"


class RPSMetric(models.RPSMetric, PatchedModelBase):
    type: Literal["rps"] = "rps"


class CronMetric(models.CronMetric, PatchedModelBase):
    type: Literal["cron"] = "cron"


class ServiceAutoscaling(models.ServiceAutoscaling, PatchedModelBase):
    pass


class AsyncServiceAutoscaling(models.AsyncServiceAutoscaling, PatchedModelBase):
    pass


class Autoscaling(ServiceAutoscaling):
    def __init__(self, **kwargs):
        warnings.warn(
            "`truefoundry.deploy.Autoscaling` is deprecated and will be removed in a future version. "
            "Please use `truefoundry.deploy.ServiceAutoscaling` instead. "
            "You can rename `Autoscaling` to `ServiceAutoscaling` in your script.",
            category=TrueFoundryDeprecationWarning,
            stacklevel=2,
        )
        super().__init__(**kwargs)


class BlueGreen(models.BlueGreen, PatchedModelBase):
    type: Literal["blue_green"] = "blue_green"


class Canary(models.Canary, PatchedModelBase):
    type: Literal["canary"] = "canary"


class Rolling(models.Rolling, PatchedModelBase):
    type: Literal["rolling_update"] = "rolling_update"


class SecretMount(models.SecretMount, PatchedModelBase):
    type: Literal["secret"] = "secret"


class StringDataMount(models.StringDataMount, PatchedModelBase):
    type: Literal["string"] = "string"


class VolumeMount(models.VolumeMount, PatchedModelBase):
    type: Literal["volume"] = "volume"


class NodeSelector(models.NodeSelector, PatchedModelBase):
    type: Literal["node_selector"] = "node_selector"


class NodepoolSelector(models.NodepoolSelector, PatchedModelBase):
    type: Literal["nodepool_selector"] = "nodepool_selector"


class Endpoint(models.Endpoint, PatchedModelBase):
    pass


class HelmRepo(models.HelmRepo, PatchedModelBase):
    type: Literal["helm-repo"] = "helm-repo"


class GitHelmRepo(models.GitHelmRepo, PatchedModelBase):
    type: Literal["git-helm-repo"] = "git-helm-repo"


class OCIRepo(models.OCIRepo, PatchedModelBase):
    type: Literal["oci-repo"] = "oci-repo"


class VolumeBrowser(models.VolumeBrowser, PatchedModelBase):
    pass


class WorkerConfig(models.WorkerConfig, PatchedModelBase):
    pass


class SQSInputConfig(models.SQSInputConfig, PatchedModelBase):
    type: Literal["sqs"] = "sqs"


class SQSOutputConfig(models.SQSOutputConfig, PatchedModelBase):
    type: Literal["sqs"] = "sqs"


class SQSQueueMetricConfig(models.SQSQueueMetricConfig, PatchedModelBase):
    type: Literal["sqs"] = "sqs"


class AWSAccessKeyAuth(models.AWSAccessKeyAuth, PatchedModelBase):
    pass


class NATSInputConfig(models.NATSInputConfig, PatchedModelBase):
    type: Literal["nats"] = "nats"


class CoreNATSOutputConfig(models.CoreNATSOutputConfig, PatchedModelBase):
    type: Literal["core-nats"] = "core-nats"


class NATSUserPasswordAuth(models.NATSUserPasswordAuth, PatchedModelBase):
    pass


class NATSOutputConfig(models.NATSOutputConfig, PatchedModelBase):
    type: Literal["nats"] = "nats"


class NATSMetricConfig(models.NATSMetricConfig, PatchedModelBase):
    type: Literal["nats"] = "nats"


class KafkaInputConfig(models.KafkaInputConfig, PatchedModelBase):
    type: Literal["kafka"] = "kafka"


class KafkaOutputConfig(models.KafkaOutputConfig, PatchedModelBase):
    type: Literal["kafka"] = "kafka"


class KafkaMetricConfig(models.KafkaMetricConfig, PatchedModelBase):
    type: Literal["kafka"] = "kafka"


class KafkaSASLAuth(models.KafkaSASLAuth, PatchedModelBase):
    pass


class AMQPInputConfig(models.AMQPInputConfig, PatchedModelBase):
    type: Literal["amqp"] = "amqp"


class AMQPOutputConfig(models.AMQPOutputConfig, PatchedModelBase):
    type: Literal["amqp"] = "amqp"


class AMQPMetricConfig(models.AMQPMetricConfig, PatchedModelBase):
    type: Literal["amqp"] = "amqp"


class AsyncProcessorSidecar(models.AsyncProcessorSidecar, PatchedModelBase):
    pass


class ArtifactsCacheVolume(models.ArtifactsCacheVolume, PatchedModelBase):
    pass


class HuggingfaceArtifactSource(models.HuggingfaceArtifactSource, PatchedModelBase):
    type: Literal["huggingface-hub"] = "huggingface-hub"


class TrueFoundryArtifactSource(models.TrueFoundryArtifactSource, PatchedModelBase):
    type: Literal["truefoundry-artifact"] = "truefoundry-artifact"


class ArtifactsDownload(models.ArtifactsDownload, PatchedModelBase):
    pass


class NvidiaGPU(models.NvidiaGPU, PatchedModelBase):
    type: Literal["nvidia_gpu"] = "nvidia_gpu"
    name: Optional[Union[GPUType, constr(regex=r"^tpu-[a-z\d\-]+$")]] = None  # type: ignore[valid-type]


class NvidiaMIGGPU(models.NvidiaMIGGPU, PatchedModelBase):
    type: Literal["nvidia_mig_gpu"] = "nvidia_mig_gpu"


class NvidiaTimeslicingGPU(models.NvidiaTimeslicingGPU, PatchedModelBase):
    type: Literal["nvidia_timeslicing_gpu"] = "nvidia_timeslicing_gpu"


class AWSInferentia(models.AWSInferentia, PatchedModelBase):
    type: Literal["aws_inferentia"] = "aws_inferentia"
    name: Optional[Union[AWSInferentiaAccelerator, str]] = None


class GcpTPU(models.GcpTPU, PatchedModelBase):
    type: Literal["gcp_tpu"] = "gcp_tpu"
    name: Union[TPUType, Literal[r"tpu-[a-z\d\-]+"]]


class DynamicVolumeConfig(models.DynamicVolumeConfig, PatchedModelBase):
    type: Literal["dynamic"] = "dynamic"


class StaticVolumeConfig(models.StaticVolumeConfig, PatchedModelBase):
    type: Literal["static"] = "static"


class PythonTaskConfig(models.PythonTaskConfig, PatchedModelBase):
    type: Literal["python-task-config"] = "python-task-config"
    resources: models.Resources = Field(default_factory=models.Resources)


class ContainerTaskConfig(models.ContainerTaskConfig, PatchedModelBase):
    type: Literal["container-task-config"] = "container-task-config"
    resources: models.Resources = Field(default_factory=models.Resources)


class TaskDockerFileBuild(models.TaskDockerFileBuild, PatchedModelBase):
    type: Literal["task-dockerfile-build"] = "task-dockerfile-build"


class TaskPythonBuild(models.TaskPythonBuild, PatchedModelBase):
    type: Literal["task-python-build"] = "task-python-build"


# Deprecated aliases, kept for backward compatibility
TruefoundryArtifactSource = TrueFoundryArtifactSource


class Email(models.Email, PatchedModelBase):
    type: Literal["email"] = "email"


class SlackWebhook(models.SlackWebhook, PatchedModelBase):
    type: Literal["slack-webhook"] = "slack-webhook"


class SlackBot(models.SlackBot, PatchedModelBase):
    type: Literal["slack-bot"] = "slack-bot"


class SparkJobScalaEntrypoint(models.SparkJobScalaEntrypoint, PatchedModelBase):
    type: Literal["scala"] = "scala"


class SparkJobPythonEntrypoint(models.SparkJobPythonEntrypoint, PatchedModelBase):
    type: Literal["python"] = "python"


class SparkJobJavaEntrypoint(models.SparkJobJavaEntrypoint, PatchedModelBase):
    type: Literal["java"] = "java"


class SparkJobPythonNotebookEntrypoint(
    models.SparkJobPythonNotebookEntrypoint, PatchedModelBase
):
    type: Literal["python-notebook"] = "python-notebook"


class SparkJobScalaNotebookEntrypoint(
    models.SparkJobScalaNotebookEntrypoint, PatchedModelBase
):
    type: Literal["scala-notebook"] = "scala-notebook"


class PySparkTaskConfig(models.PySparkTaskConfig, PatchedModelBase):
    type: Literal["pyspark-task-config"] = "pyspark-task-config"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            import truefoundry.workflow.spark_task as _  # noqa: F401
        except ImportError as e:
            raise ImportError(
                "truefoundry.workflow.spark_task is not installed. Please install it with `pip install truefoundry[workflow,spark]`"
            ) from e


class DatabricksJobTaskConfig(models.DatabricksJobTaskConfig, PatchedModelBase):
    type: Literal["databricks-job-task-config"] = "databricks-job-task-config"
    resources: models.Resources = Field(default_factory=models.Resources)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            import truefoundry.workflow.databricks_task as _  # noqa: F401
        except ImportError as e:
            raise ImportError(
                "truefoundry.workflow.databricks_task is not installed. Please install it with `pip install truefoundry[workflow]`"
            ) from e


class SparkDriverConfig(models.SparkDriverConfig, PatchedModelBase):
    type: Literal["spark-driver-config"] = "spark-driver-config"


class SparkExecutorConfig(models.SparkExecutorConfig, PatchedModelBase):
    type: Literal["spark-executor-config"] = "spark-executor-config"


class SparkExecutorFixedInstances(models.SparkExecutorFixedInstances, PatchedModelBase):
    type: Literal["fixed"] = "fixed"


class SparkExecutorDynamicScaling(models.SparkExecutorDynamicScaling, PatchedModelBase):
    type: Literal["dynamic"] = "dynamic"


class TaskPySparkBuild(models.TaskPySparkBuild, PatchedModelBase):
    type: Literal["task-pyspark-build"] = "task-pyspark-build"


class Pip(models.Pip, PatchedModelBase):
    type: Literal["pip"] = "pip"


class UV(models.UV, PatchedModelBase):
    type: Literal["uv"] = "uv"


class Poetry(models.Poetry, PatchedModelBase):
    type: Literal["poetry"] = "poetry"


class AwsSqsAccessKeyBasedAuth(models.AwsSqsAccessKeyBasedAuth, PatchedModelBase):
    type: Literal["access-key-based"] = "access-key-based"


class AwsSqsAssumedRoleBasedAuth(models.AwsSqsAssumedRoleBasedAuth, PatchedModelBase):
    type: Literal["assumed-role-based"] = "assumed-role-based"
